import logging
import pickle

import numpy as np
import pkg_resources
import pytest

import nengo
import nengo.simulator
from nengo.builder import Model
from nengo.builder.ensemble import BuiltEnsemble
from nengo.builder.operator import DotInc
from nengo.builder.signal import Signal
from nengo.exceptions import ReadonlyError, SimulatorClosed, ValidationError
from nengo.rc import RC_DEFAULTS, rc
from nengo.utils.progress import ProgressBar


def test_steps(Simulator, allclose):
    """Tests stepping through a simple simulation, ensuring that
    steps are tracked and steps take the right amount of time"""
    dt = 0.001
    m = nengo.Network(label="test_steps")
    with Simulator(m, dt=dt) as sim:
        assert sim.n_steps == 0
        assert allclose(sim.time, 0 * dt)
        sim.step()
        assert sim.n_steps == 1
        assert allclose(sim.time, 1 * dt)
        sim.step()
        assert sim.n_steps == 2
        assert allclose(sim.time, 2 * dt)

        assert np.isscalar(sim.n_steps)
        assert np.isscalar(sim.time)


@pytest.mark.parametrize("bits", ["16", "32", "64"])
def test_dtype(Simulator, request, seed, bits):
    # Ensures dtype is set back to default after the test, even if it fails
    request.addfinalizer(
        lambda: rc.set("precision", "bits", str(RC_DEFAULTS["precision"]["bits"]))
    )

    float_dtype = np.dtype(getattr(np, f"float{bits}"))
    int_dtype = np.dtype(getattr(np, f"int{bits}"))

    with nengo.Network() as model:
        u = nengo.Node([0.5, -0.4])
        a = nengo.Ensemble(10, 2)
        nengo.Connection(u, a)
        p = nengo.Probe(a)

    rc["precision"]["bits"] = bits
    with Simulator(model) as sim:
        sim.step()

        for k, v in sim.signals.items():
            assert v.dtype in (float_dtype, int_dtype), f"Signal '{k}' wrong dtype"

        objs = (obj for obj in model.all_objects if sim.data[obj] is not None)
        for obj in objs:
            for x in (x for x in sim.data[obj] if isinstance(x, np.ndarray)):
                assert x.dtype == float_dtype, obj

        assert sim.data[p].dtype == float_dtype


def test_time_absolute(Simulator, allclose):
    m = nengo.Network()
    with Simulator(m) as sim:
        sim.run(0.003)
    assert allclose(sim.trange(), [0.001, 0.002, 0.003])


def test_trange_with_probes(Simulator):
    dt = 1e-3
    m = nengo.Network()
    periods = dt * np.arange(1, 21)
    with m:
        u = nengo.Node(output=np.sin)
        probes = [nengo.Probe(u, sample_every=p, synapse=5 * p) for p in periods]

    with Simulator(m, dt=dt) as sim:
        sim.run(0.333)
    for i, p in enumerate(periods):
        assert len(sim.trange(sample_every=p)) == len(sim.data[probes[i]])


def test_simulation_data():
    """Tests simulator.SimulationData's implementation."""
    raw = {"scalar": 5, "list": [2, 4, 6]}
    data = nengo.simulator.SimulationData(raw)
    assert np.all(data["scalar"] == np.asarray(raw["scalar"]))
    assert np.all(data.get("list") == np.asarray(raw.get("list")))
    assert tuple(data) == tuple(raw)  # this tests __iter__
    assert len(data) == len(raw)
    assert repr(data) == repr(raw)
    assert str(data) == str(raw)

    assert len(data._cache) > 0
    data.reset()
    assert len(data._cache) == 0


def test_simulation_data_with_repeated_simulator_runs(Simulator):
    with nengo.Network() as model:
        ens = nengo.Ensemble(10, 1)
        p = nengo.Probe(ens)

    dt = 0.001
    with Simulator(model, dt=dt) as sim:
        sim.run(0.01)
        assert len(sim.data[p]) == 10
        sim.run(0.01)
        assert len(sim.data[p]) == 20


def test_close_function(Simulator):
    m = nengo.Network()
    with m:
        nengo.Ensemble(10, 1)

    sim = Simulator(m)
    sim.close()
    with pytest.raises(SimulatorClosed):
        sim.run(1.0)
    with pytest.raises(SimulatorClosed):
        sim.reset()


def test_close_context(Simulator):
    m = nengo.Network()
    with m:
        nengo.Ensemble(10, 1)

    with Simulator(m) as sim:
        sim.run(0.01)

    with pytest.raises(SimulatorClosed):
        sim.run(1.0)
    with pytest.raises(SimulatorClosed):
        sim.reset()


def test_close_steps(Simulator):
    """For Simulator, closed simulators should fail for ``step``"""
    m = nengo.Network()
    with m:
        nengo.Ensemble(10, 1)

    # test close function
    sim = Simulator(m)
    sim.close()
    with pytest.raises(SimulatorClosed):
        sim.run_steps(1)
    with pytest.raises(SimulatorClosed):
        sim.step()

    # test close context
    with Simulator(m) as sim:
        sim.run(0.01)

    with pytest.raises(SimulatorClosed):
        sim.run_steps(1)
    with pytest.raises(SimulatorClosed):
        sim.step()


def test_sim_reopen(Simulator):
    with Simulator(nengo.Network()) as sim:
        assert not sim.closed

    assert sim.closed

    with pytest.raises(SimulatorClosed, match="simulator is closed"):
        with sim:
            pass


def test_warn_on_opensim_del(Simulator):
    with nengo.Network() as net:
        nengo.Ensemble(10, 1)

    sim = Simulator(net)
    with pytest.warns(ResourceWarning, match="Simulator.*deallocated while open"):
        sim.__del__()
    sim.close()


def test_entry_point(Simulator):
    sims = [ep.load() for ep in pkg_resources.iter_entry_points(group="nengo.backends")]
    assert Simulator in sims


def test_signal_init_values(Simulator):
    """Tests that initial values are not overwritten."""
    zero = Signal([0])
    one = Signal([1])
    five = Signal([5.0])
    zeroarray = Signal([[0], [0], [0]])
    array = Signal([1, 2, 3])

    m = Model(dt=0)
    m.operators += [DotInc(zero, zero, five), DotInc(zeroarray, one, array)]

    with Simulator(None, model=m) as sim:
        assert sim.signals[zero][0] == 0
        assert sim.signals[one][0] == 1
        assert sim.signals[five][0] == 5.0
        assert np.all(np.array([1, 2, 3]) == sim.signals[array])
        sim.step()
        assert sim.signals[zero][0] == 0
        assert sim.signals[one][0] == 1
        assert sim.signals[five][0] == 5.0
        assert np.all(np.array([1, 2, 3]) == sim.signals[array])


def test_seeding(Simulator, allclose):
    """Test that setting the model seed fixes everything"""

    #  TODO: this really just checks random parameters in ensembles.
    #   Are there other objects with random parameters that should be
    #   tested? (Perhaps initial weights of learned connections)

    m = nengo.Network(label="test_seeding")
    with m:
        input = nengo.Node(output=1, label="input")
        A = nengo.Ensemble(40, 1, label="A")
        B = nengo.Ensemble(20, 1, label="B")
        nengo.Connection(input, A)
        C = nengo.Connection(A, B, function=lambda x: x ** 2)

    m.seed = 872
    with Simulator(m) as sim:
        m1 = sim.data
    with Simulator(m) as sim:
        m2 = sim.data
    m.seed = 873
    with Simulator(m) as sim:
        m3 = sim.data

    def compare_objs(obj1, obj2, attrs, equal=True):
        for attr in attrs:
            attr1 = getattr(obj1, attr)
            attr2 = getattr(obj2, attr)
            check = (
                allclose(attr1, attr2)
                if equal
                else (not allclose(attr1, attr2, record_rmse=False, print_fail=0))
            )
            if not check:
                logging.info("%s: %s", attr, attr1)
                logging.info("%s: %s", attr, attr2)
            assert check

    ens_attrs = BuiltEnsemble._fields
    As = [mi[A] for mi in [m1, m2, m3]]
    Bs = [mi[B] for mi in [m1, m2, m3]]
    compare_objs(As[0], As[1], ens_attrs)
    compare_objs(Bs[0], Bs[1], ens_attrs)
    compare_objs(As[0], As[2], ens_attrs, equal=False)
    compare_objs(Bs[0], Bs[2], ens_attrs, equal=False)

    conn_attrs = ("eval_points", "weights")
    Cs = [mi[C] for mi in [m1, m2, m3]]
    compare_objs(Cs[0], Cs[1], conn_attrs)
    compare_objs(Cs[0], Cs[2], conn_attrs, equal=False)


def test_hierarchical_seeding():
    """Changes to subnetworks shouldn't affect seeds in top-level network"""

    def create(make_extra, seed):
        objs = []
        with nengo.Network(seed=seed, label="n1") as model:
            objs.append(nengo.Ensemble(10, 1, label="e1"))
            with nengo.Network(label="n2"):
                objs.append(nengo.Ensemble(10, 1, label="e2"))
                if make_extra:
                    # This shouldn't affect any seeds
                    objs.append(nengo.Ensemble(10, 1, label="e3"))
            objs.append(nengo.Ensemble(10, 1, label="e4"))
        return model, objs

    same1, same1objs = create(False, 9)
    same2, same2objs = create(True, 9)
    diff, diffobjs = create(True, 10)

    m1 = Model()
    m1.build(same1)
    same1seeds = m1.seeds

    m2 = Model()
    m2.build(same2)
    same2seeds = m2.seeds

    m3 = Model()
    m3.build(diff)
    diffseeds = m3.seeds

    for diffobj, same2obj in zip(diffobjs, same2objs):
        # These seeds should all be different
        assert diffseeds[diffobj] != same2seeds[same2obj]

    # Skip the extra ensemble
    same2objs = same2objs[:2] + same2objs[3:]

    for same1obj, same2obj in zip(same1objs, same2objs):
        # These seeds should all be the same
        assert same1seeds[same1obj] == same2seeds[same2obj]


def test_probe_cache(Simulator, allclose):
    with nengo.Network() as model:
        u = nengo.Node(nengo.processes.WhiteNoise())
        up = nengo.Probe(u)

    with Simulator(model, seed=0) as sim:
        sim.run_steps(10)
        ua = np.array(sim.data[up])

        sim.reset(seed=1)
        sim.run_steps(10)
        ub = np.array(sim.data[up])

    assert not allclose(ua, ub, atol=1e-1, record_rmse=False, print_fail=0)


def test_invalid_run_time(Simulator):
    net = nengo.Network()
    with Simulator(net) as sim:
        with pytest.raises(ValidationError):
            sim.run(-0.0001)
        with pytest.warns(UserWarning):
            sim.run(0)
        sim.run(0.0006)  # Rounds up to 0.001
        assert sim.n_steps == 1


def test_sim_seed_set_by_network_seed(Simulator, seed):
    with nengo.Network(seed=seed) as model:
        pass
    with Simulator(model) as sim:
        sim_seed = sim.seed
    with Simulator(model) as sim:
        assert sim.seed == sim_seed


def test_simulator_progress_bars(Simulator):
    class ProgressBarInvariants(ProgressBar):
        def __init__(self):
            super().__init__()
            self.progress = None
            self.max_steps = None
            self.n_steps = 0
            self.closed = False

        def update(self, progress):
            assert not self.closed
            if self.progress is not progress:
                assert self.max_steps is None or self.n_steps <= self.max_steps
                self.n_steps = progress.n_steps
                self.max_steps = progress.max_steps
                self.progress = progress
            assert progress.max_steps == self.max_steps
            assert self.max_steps is None or self.n_steps <= progress.n_steps
            self.n_steps = progress.n_steps

        def close(self):
            self.closed = True

    with nengo.Network() as model:
        for _ in range(3):
            for _ in range(3):
                nengo.Ensemble(10, 1)
            with nengo.Network():
                for _ in range(3):
                    nengo.Ensemble(10, 1)
    build_invariants = ProgressBarInvariants()
    with Simulator(model, progress_bar=build_invariants) as sim:
        run_invariants = ProgressBarInvariants()
        sim.run(0.01, progress_bar=run_invariants)
        assert run_invariants.n_steps == run_invariants.max_steps


@pytest.mark.parametrize("sample_every", (0.001, 0.0005, 0.002, 0.0015))
def test_sample_every_trange(Simulator, sample_every, allclose):
    with nengo.Network() as model:
        t = nengo.Node(lambda t: t)
        p = nengo.Probe(t, sample_every=sample_every)

    with Simulator(model) as sim:
        sim.run(0.01)

    with pytest.raises(ValidationError):
        sim.trange(dt=sample_every, sample_every=sample_every)
    with pytest.warns(DeprecationWarning):
        assert allclose(sim.trange(dt=sample_every), np.squeeze(sim.data[p]))
    assert allclose(sim.trange(sample_every=sample_every), np.squeeze(sim.data[p]))


def test_pickle_optimize(caplog, seed):
    caplog.set_level(logging.DEBUG)
    with nengo.Network(seed=seed) as net:
        stim = nengo.Node(np.zeros(10))
        ea = nengo.networks.EnsembleArray(20, stim.size_out)
        nengo.Connection(stim, ea.input)
        net.probe = nengo.Probe(ea.output)

    with nengo.Simulator(net, optimize=True) as sim:
        pickled = pickle.dumps(sim)
        sim.run(0.01)
        assert not sim.closed
        assert sim.optimize

    before = sim.data[net.probe]
    del net, sim

    # Check that original sim was optimized
    assert any(record.msg == "Optimizing model..." for record in caplog.records)
    caplog.clear()

    unpickled = pickle.loads(pickled)
    net = unpickled.model.toplevel
    assert not unpickled.closed
    assert unpickled.optimize

    # Check that unpickling doesn't re-do optimization
    assert not any(record.msg == "Optimizing model..." for record in caplog.records)

    unpickled.run(0.01)
    after = unpickled.data[net.probe]
    unpickled.close()

    assert np.all(before == after)


def test_dt_readonly():
    with nengo.Simulator(nengo.Network()) as sim:
        with pytest.raises(ReadonlyError, match="dt"):
            sim.dt = 0.05



from collections import namedtuple

import numpy as np

from nengo.builder.builder import Builder
from nengo.builder.ensemble import gen_eval_points, get_activities
from nengo.builder.node import SimPyFunc
from nengo.builder.operator import Copy, ElementwiseInc, Reset
from nengo.builder.signal import Signal
from nengo.connection import Connection
from nengo.ensemble import Ensemble, Neurons
from nengo.exceptions import BuildError
from nengo.neurons import Direct
from nengo.node import Node
from nengo.rc import rc
from nengo.solvers import NoSolver, Solver
from nengo.transforms import Dense, NoTransform
from nengo.utils.numpy import is_integer, is_iterable

built_attrs = ["eval_points", "solver_info", "weights", "transform"]


class BuiltConnection(namedtuple("BuiltConnection", built_attrs)):
    """Collects the parameters generated in `.build_connection`.

    These are stored here because in the majority of cases the equivalent
    attribute in the original connection is a `.Distribution`. The attributes
    of a BuiltConnection are the full NumPy arrays used in the simulation.

    See the `.Connection` documentation for more details on each parameter.

    Parameters
    ----------
    eval_points : ndarray
        Evaluation points.
    solver_info : dict
        Information dictionary returned by the `.Solver`.
    weights : ndarray
        Connection weights. May be synaptic connection weights defined in
        the connection's transform, or a combination of the decoders
        automatically solved for and the specified transform.
    transform : ndarray
        The transform matrix.
    """

    __slots__ = ()

    def __new__(cls, eval_points, solver_info, weights, transform):
        # Overridden to suppress the default __new__ docstring
        return tuple.__new__(cls, (eval_points, solver_info, weights, transform))


def get_eval_points(model, conn, rng):
    """Get evaluation points for connection."""

    if conn.eval_points is None:
        view = model.params[conn.pre_obj].eval_points.view()
        view.setflags(write=False)
        assert view.dtype == rc.float_dtype
        return view
    else:
        return gen_eval_points(
            conn.pre_obj,
            conn.eval_points,
            rng,
            conn.scale_eval_points,
            dtype=rc.float_dtype,
        )


def get_targets(conn, eval_points, dtype=None):
    """Get target points for connection with given evaluation points."""

    dtype = rc.float_dtype if dtype is None else dtype
    if conn.function is None:
        targets = eval_points[:, conn.pre_slice].astype(dtype)
    elif isinstance(conn.function, np.ndarray):
        targets = conn.function
    else:
        targets = np.zeros((len(eval_points), conn.size_mid), dtype=dtype)
        for i, ep in enumerate(eval_points[:, conn.pre_slice]):
            out = conn.function(ep)
            if out is None:
                raise BuildError(
                    f"Building {conn}: Connection function returned "
                    "None. Cannot solve for decoders."
                )
            if isinstance(out, (tuple, list)):
                out = np.hstack(out)
            targets[i] = out

    return targets


def build_linear_system(model, conn, rng):
    """Get all arrays needed to compute decoders."""
    eval_points = get_eval_points(model, conn, rng)
    ens = conn.pre_obj
    activities = get_activities(model.params[ens], ens, eval_points)
    if np.count_nonzero(activities) == 0:
        raise BuildError(
            f"Building {conn}: 'activities' matrix is all zero for {conn.pre_obj}. "
            "This is because no evaluation points fall in the firing "
            "ranges of any neurons."
        )

    targets = get_targets(conn, eval_points, dtype=rc.float_dtype)
    return eval_points, activities, targets


def build_decoders(model, conn, rng):
    """Compute decoders for connection."""

    encoders = model.params[conn.pre_obj].encoders
    gain = model.params[conn.pre_obj].gain
    bias = model.params[conn.pre_obj].bias

    eval_points = get_eval_points(model, conn, rng)
    targets = get_targets(conn, eval_points, dtype=rc.float_dtype)

    if conn.solver.weights and not conn.solver.compositional:
        # solver is solving for the whole weight matrix, so apply
        # transform/encoders to targets
        if not isinstance(conn.transform, Dense):
            raise BuildError(
                "Non-compositional solvers only work with Dense transforms"
            )
        transform = conn.transform.sample(rng=rng)
        targets = np.dot(targets, transform.T)
        # weight solvers only allowed on ensemble->ensemble connections
        assert isinstance(conn.post_obj, Ensemble)
        post_enc = model.params[conn.post_obj].scaled_encoders
        targets = np.dot(targets, post_enc.T[conn.post_slice])

    x = np.dot(eval_points, encoders.T / conn.pre_obj.radius)
    wrapped_solver = (
        model.decoder_cache.wrap_solver(solve_for_decoders)
        if model.seeded[conn]
        else solve_for_decoders
    )
    decoders, solver_info = wrapped_solver(conn, gain, bias, x, targets, rng=rng)

    return eval_points, decoders.T, solver_info


def solve_for_decoders(conn, gain, bias, x, targets, rng):
    """Solver for decoders.

    Factored out from `.build_decoders` for use with the cache system.
    """

    activities = conn.pre_obj.neuron_type.rates(x, gain, bias)
    if np.count_nonzero(activities) == 0:
        raise BuildError(
            f"Building {conn}: 'activities' matrix is all zero for {conn.pre_obj}. "
            "This is because no evaluation points fall in the firing "
            "ranges of any neurons."
        )

    decoders, solver_info = conn.solver(activities, targets, rng=rng)
    return decoders, solver_info


def slice_signal(model, signal, sl):
    """Apply a slice operation to given signal."""
    assert signal.ndim == 1
    if isinstance(sl, slice) and (sl.step is None or sl.step == 1):
        return signal[sl]
    else:
        size = np.arange(signal.size, dtype=rc.float_dtype)[sl].size
        sliced_signal = Signal(shape=size, name=f"{signal.name}.sliced")
        model.add_op(Copy(signal, sliced_signal, src_slice=sl))
        return sliced_signal


@Builder.register(Solver)
def build_solver(model, solver, conn, rng):
    """Apply decoder solver to connection."""
    return build_decoders(model, conn, rng)


@Builder.register(NoSolver)
def build_no_solver(model, solver, conn, rng):
    """Special builder for NoSolver to skip unnecessary steps."""
    activities = np.zeros((1, conn.pre_obj.n_neurons), dtype=rc.float_dtype)
    targets = np.zeros((1, conn.size_mid), dtype=rc.float_dtype)
    # No need to invoke the cache for NoSolver
    decoders, solver_info = conn.solver(activities, targets, rng=rng)
    weights = decoders.T
    return None, weights, solver_info


@Builder.register(Connection)  # noqa: C901
def build_connection(model, conn):  # noqa: C901
    """Builds a `.Connection` object into a model.

    A brief summary of what happens in the connection build process,
    in order:

    1. Solve for decoders.
    2. Combine transform matrix with decoders to get weights.
    3. Add operators for computing the function
       or multiplying neural activity by weights.
    4. Call build function for the synapse.
    5. Call build function for the learning rule.
    6. Add operator for applying learning rule delta to weights.

    Some of these steps may be altered or omitted depending on the parameters
    of the connection, in particular the pre and post types.

    Parameters
    ----------
    model : Model
        The model to build into.
    conn : Connection
        The connection to build.

    Notes
    -----
    Sets ``model.params[conn]`` to a `.BuiltConnection` instance.
    """

    # Create random number generator
    rng = np.random.RandomState(model.seeds[conn])

    # Get input and output connections from pre and post
    def get_prepost_signal(is_pre):
        target = conn.pre_obj if is_pre else conn.post_obj
        key = "out" if is_pre else "in"

        if target not in model.sig:
            raise BuildError(
                f"Building {conn}: the '{'pre' if is_pre else 'post'}' object {target} "
                "is not in the model, or has a size of zero."
            )
        signal = model.sig[target].get(key, None)
        if signal is None or signal.size == 0:
            raise BuildError(
                f"Building {conn}: the '{'pre' if is_pre else 'post'}' object {target} "
                f"has a '{key}' size of zero."
            )

        return signal

    model.sig[conn]["in"] = get_prepost_signal(is_pre=True)
    model.sig[conn]["out"] = get_prepost_signal(is_pre=False)

    decoders = None
    encoders = None
    eval_points = None
    solver_info = None
    post_slice = conn.post_slice

    # Figure out the signal going across this connection
    in_signal = model.sig[conn]["in"]
    if isinstance(conn.pre_obj, Node) or (
        isinstance(conn.pre_obj, Ensemble)
        and isinstance(conn.pre_obj.neuron_type, Direct)
    ):
        # Node or Decoded connection in directmode
        sliced_in = slice_signal(model, in_signal, conn.pre_slice)
        if conn.function is None:
            in_signal = sliced_in
        elif isinstance(conn.function, np.ndarray):
            raise BuildError("Cannot use function points in direct connection")
        else:
            in_signal = Signal(shape=conn.size_mid, name=f"{conn}.func")
            model.add_op(SimPyFunc(in_signal, conn.function, None, sliced_in))
    elif isinstance(conn.pre_obj, Ensemble):  # Normal decoded connection
        eval_points, decoders, solver_info = model.build(conn.solver, conn, rng)
        if isinstance(conn.post_obj, Ensemble) and conn.solver.weights:
            model.sig[conn]["out"] = model.sig[conn.post_obj.neurons]["in"]

            encoders = model.params[conn.post_obj].scaled_encoders.T
            encoders = encoders[conn.post_slice]

            # post slice already applied to encoders (either here or in
            # `build_decoders`), so don't apply later
            post_slice = None
    else:
        in_signal = slice_signal(model, in_signal, conn.pre_slice)

    # Build transform
    if conn.solver.weights and not conn.solver.compositional:
        # special case for non-compositional weight solvers, where
        # the solver is solving for the full weight matrix. so we don't
        # need to combine decoders/transform/encoders.
        weighted, weights = model.build(
            Dense(decoders.shape, init=decoders), in_signal, rng=rng
        )
    else:
        weighted, weights = model.build(
            conn.transform, in_signal, decoders=decoders, encoders=encoders, rng=rng
        )

    model.sig[conn]["weights"] = weights

    # Build synapse
    if conn.synapse is not None:
        weighted = model.build(conn.synapse, weighted, mode="update")

    # Store the weighted-filtered output in case we want to probe it
    model.sig[conn]["weighted"] = weighted

    if isinstance(conn.post_obj, Neurons):
        # Apply neuron gains (we don't need to do this if we're connecting to
        # an Ensemble, because the gains are rolled into the encoders)
        gains = Signal(
            model.params[conn.post_obj.ensemble].gain[post_slice],
            name=f"{conn}.gains",
        )

        if is_integer(post_slice) or isinstance(post_slice, slice):
            sliced_out = model.sig[conn]["out"][post_slice]
        else:
            # advanced indexing not supported on Signals, so we need to set up an
            # intermediate signal and use a Copy op to perform the indexing
            sliced_out = Signal(shape=gains.shape, name=f"{conn}.sliced_out")
            model.add_op(Reset(sliced_out))
            model.add_op(
                Copy(sliced_out, model.sig[conn]["out"], dst_slice=post_slice, inc=True)
            )

        model.add_op(
            ElementwiseInc(
                gains, weighted, sliced_out, tag=f"{conn}.gains_elementwiseinc"
            )
        )
    else:
        # Copy to the proper slice
        model.add_op(
            Copy(
                weighted,
                model.sig[conn]["out"],
                dst_slice=post_slice,
                inc=True,
                tag=f"{conn}",
            )
        )

    # Build learning rules
    if conn.learning_rule is not None:
        # TODO: provide a general way for transforms to expose learnable params
        if not isinstance(conn.transform, (Dense, NoTransform)):
            raise NotImplementedError(
                f"Learning on connections with {type(conn.transform).__name__} "
                "transforms is not supported"
            )

        rule = conn.learning_rule
        rule = [rule] if not is_iterable(rule) else rule
        targets = []
        for r in rule.values() if isinstance(rule, dict) else rule:
            model.build(r)
            targets.append(r.modifies)

        if "encoders" in targets:
            encoder_sig = model.sig[conn.post_obj]["encoders"]
            encoder_sig.readonly = False
        if "decoders" in targets or "weights" in targets:
            if weights.ndim < 2:
                raise BuildError(
                    "'transform' must be a 2-dimensional array for learning"
                )
            model.sig[conn]["weights"].readonly = False

    model.params[conn] = BuiltConnection(
        eval_points=eval_points,
        solver_info=solver_info,
        transform=conn.transform,
        weights=getattr(weights, "initial_value", None),
    )



from nengo.builder.builder import Builder
from nengo.builder.operator import Reset, SimPyFunc
from nengo.builder.signal import Signal
from nengo.exceptions import BuildError
from nengo.node import Node
from nengo.processes import Process
from nengo.rc import rc
from nengo.utils.numpy import is_array_like


@Builder.register(Node)
def build_node(model, node):
    """Builds a `.Node` object into a model.

    The node build function is relatively simple. It involves creating input
    and output signals, and connecting them with an `.Operator` that depends
    on the type of ``node.output``.

    Parameters
    ----------
    model : Model
        The model to build into.
    node : Node
        The node to build.

    Notes
    -----
    Sets ``model.params[node]`` to ``None``.
    """

    # input signal
    if not is_array_like(node.output) and node.size_in > 0:
        sig_in = Signal(shape=node.size_in, name=f"{node}.in")
        model.add_op(Reset(sig_in))
    else:
        sig_in = None

    # Provide output
    if node.output is None:
        sig_out = sig_in
    elif isinstance(node.output, Process):
        sig_out = Signal(shape=node.size_out, name=f"{node}.out")
        model.build(node.output, sig_in, sig_out, mode="set")
    elif callable(node.output):
        sig_out = (
            Signal(shape=node.size_out, name=f"{node}.out")
            if node.size_out > 0
            else None
        )
        model.add_op(SimPyFunc(output=sig_out, fn=node.output, t=model.time, x=sig_in))
    elif is_array_like(node.output):
        sig_out = Signal(node.output.astype(rc.float_dtype), name=f"{node}.out")
    else:
        raise BuildError(f"Invalid node output type '{type(node.output).__name__}'")

    model.sig[node]["in"] = sig_in
    model.sig[node]["out"] = sig_out
    model.params[node] = None



import _pytest.capture
import pytest

from nengo.utils.paths import examples_dir
from nengo.utils.stdlib import execfile

try:
    from nengo.utils.ipython import export_py, iter_cells, load_notebook
except ImportError as err:

    def export_py(*args, import_err=err, **kwargs):
        raise import_err

    def load_notebook(*args, import_err=err, **kwargs):
        raise import_err


# Monkeypatch _pytest.capture.DontReadFromInput
#  If we don't do this, importing IPython will choke as it reads the current
#  sys.stdin to figure out the encoding it will use; pytest installs
#  DontReadFromInput as sys.stdin to capture output.
#  Running with -s option doesn't have this issue, but this monkeypatch
#  doesn't have any side effects, so it's fine.
_pytest.capture.DontReadFromInput.encoding = "utf-8"
_pytest.capture.DontReadFromInput.write = lambda: None
_pytest.capture.DontReadFromInput.flush = lambda: None


too_slow = {
    "advanced/inhibitory-gating",
    "advanced/izhikevich",
    "advanced/nef-algorithm",
    "advanced/nef-summary",
    "dynamics/lorenz-attractor",
    "learning/learn-communication-channel",
    "learning/learn-product",
    "learning/learn-square",
    "learning/learn-unsupervised",
    "learning/lmu",
    "networks/basal-ganglia",
    "usage/network-design",
    "usage/network-design-advanced",
}

files = [
    f for f in examples_dir.glob("**/*.ipynb") if ".ipynb_checkpoints" not in str(f)
]
all_examples = [f.relative_to(examples_dir).with_suffix("") for f in files]
slow_examples = [f for f in all_examples if f.as_posix() in too_slow]
fast_examples = [f for f in all_examples if f.as_posix() not in too_slow]

# glob goes in arbitrary order, so sort after the fact to keep pytest happy
# convert paths to literal strings so that they're displayed nicer in the pytest name
all_examples = sorted(str(e) for e in all_examples)
slow_examples = sorted(str(e) for e in slow_examples)
fast_examples = sorted(str(e) for e in fast_examples)


def assert_noexceptions(nb_file, tmp_path):
    plt = pytest.importorskip("matplotlib.pyplot")
    nb_path = examples_dir / f"{nb_file}.ipynb"
    nb = load_notebook(nb_path)
    pyfile = tmp_path / (nb_path.stem + ".py")
    export_py(nb, pyfile)
    execfile(pyfile, {})
    plt.close("all")


@pytest.mark.example
@pytest.mark.parametrize("nb_file", fast_examples)
@pytest.mark.filterwarnings("ignore:Creating new attribute 'memory_location'")
@pytest.mark.filterwarnings("ignore:Matplotlib is currently using agg")
def test_fast_noexceptions(nb_file, tmp_path):
    """Ensure that no cells raise an exception."""
    pytest.importorskip("IPython", minversion="3.0")
    pytest.importorskip("jinja2")
    assert_noexceptions(nb_file, tmp_path)


@pytest.mark.slow
@pytest.mark.example
@pytest.mark.parametrize("nb_file", slow_examples)
def test_slow_noexceptions(nb_file, tmp_path):
    """Ensure that no cells raise an exception."""
    pytest.importorskip("IPython", minversion="3.0")
    pytest.importorskip("jinja2")
    assert_noexceptions(nb_file, tmp_path)


@pytest.mark.example
@pytest.mark.parametrize("nb_file", all_examples)
def test_no_outputs(nb_file):
    """Ensure that no cells have output."""
    pytest.importorskip("IPython", minversion="3.0")
    nb = load_notebook(examples_dir / f"{nb_file}.ipynb")
    for cell in iter_cells(nb):
        assert cell.outputs == [], "Cell outputs not cleared"
        assert cell.execution_count is None, "Execution count not cleared"


@pytest.mark.example
@pytest.mark.parametrize("nb_file", all_examples)
def test_version_4(nb_file):
    pytest.importorskip("IPython", minversion="3.0")
    nb = load_notebook(examples_dir / f"{nb_file}.ipynb")
    assert nb.nbformat == 4


@pytest.mark.example
@pytest.mark.parametrize("nb_file", all_examples)
def test_minimal_metadata(nb_file):
    pytest.importorskip("IPython", minversion="3.0")
    nb = load_notebook(examples_dir / f"{nb_file}.ipynb")

    assert "kernelspec" not in nb.metadata
    assert "signature" not in nb.metadata

    badinfo = (
        "codemirror_mode",
        "file_extension",
        "mimetype",
        "nbconvert_exporter",
        "version",
    )
    for info in badinfo:
        assert info not in nb.metadata.language_info



from nengo.exceptions import MovedError


def spikes2events(*args, **kwargs):
    """Moved to nengo_extras.neurons."""
    raise MovedError(location="nengo_extras.neurons")


def _rates_isi_events(*args, **kwargs):
    """Moved to nengo_extras.neurons."""
    raise MovedError(location="nengo_extras.neurons")


def rates_isi(*args, **kwargs):
    """Moved to nengo_extras.neurons."""
    raise MovedError(location="nengo_extras.neurons")


def lowpass_filter(*args, **kwargs):
    """Moved to nengo_extras.neurons."""
    raise MovedError(location="nengo_extras.neurons")


def rates_kernel(*args, **kwargs):
    """Moved to nengo_extras.neurons."""
    raise MovedError(location="nengo_extras.neurons")


def settled_firingrate(*args, **kwargs):
    """Moved to nengo.neurons."""
    raise MovedError(location="nengo.neurons")



import numpy as np
import pytest

from nengo.builder import Signal
from nengo.builder.operator import (
    Copy,
    DotInc,
    ElementwiseInc,
    Operator,
    Reset,
    SimPyFunc,
    SparseDotInc,
    TimeUpdate,
    reshape_dot,
)
from nengo.exceptions import BuildError


def _test_operator_arg_attributes(OperatorType, argnames, args=None, non_signals=()):
    args = {} if args is None else args
    for argname in argnames:
        args.setdefault(argname, argname)

    sim = OperatorType(**args)
    for argname in argnames:
        assert getattr(sim, argname) == args[argname]

    signals = set()
    signals.update(sim.sets)
    signals.update(sim.incs)
    signals.update(sim.reads)
    signals.update(sim.updates)
    for argname in argnames:
        if argname in non_signals:
            continue

        signal = args[argname]
        assert signal in signals, f"{argname!r} not added to sets/incs/reads/updates"
        signals.remove(signal)

    assert len(signals) == 0, f"Extra signals in sets/incs/reads/updates: {signals!r}"

    return args, sim


def test_operator_str_repr():
    assert str(Operator(tag="hi")) == "Operator{ 'hi'}"
    assert repr(Operator(tag="hi")).startswith("<Operator 'hi' at ")


def test_timeupdate_op():
    argnames = ["step", "time"]
    _, sim = _test_operator_arg_attributes(TimeUpdate, argnames)
    assert str(sim) == "TimeUpdate{}"


def test_reset_op():
    argnames = ["dst"]
    _, sim = _test_operator_arg_attributes(Reset, argnames, args={"dst": "dval"})
    assert str(sim) == "Reset{dval}"


def test_copy_op(rng):
    argnames = ["src", "dst"]
    args = {"src": "sval", "dst": "dval"}
    _, sim = _test_operator_arg_attributes(Copy, argnames, args=args)
    assert str(sim) == "Copy{sval -> dval, inc=False}"

    cp = Copy(1, 2)
    signals = [1, 2, [1, 2, 3]]
    dt = 0
    cp.dst_slice = np.array([1, 1, 3])  # repeat 1
    with pytest.raises(BuildError, match="Cannot have repeated indices"):
        cp.make_step(signals, dt, rng)


def test_elementwiseinc_op(rng):
    argnames = ["A", "X", "Y"]
    args = {"A": "Av", "X": "Xv", "Y": "Yv"}
    _, sim = _test_operator_arg_attributes(ElementwiseInc, argnames, args=args)
    assert str(sim) == "ElementwiseInc{Av, Xv -> Yv}"

    ew = ElementwiseInc(0, 1, 2)
    signals = [np.array([1]), np.array([2]), np.array([5, 6])]
    dt = 0
    with pytest.raises(BuildError, match="Incompatible shapes in ElementwiseInc"):
        ew.make_step(signals, dt, rng)


def test_reshape_dot(rng):
    scalar = np.array(1)
    vec = [np.ones(i) for i in range(4)]
    mat11 = np.ones((1, 1))
    mat23 = np.ones((2, 3))
    mat33 = np.ones((3, 3))

    # if A.shape == ():
    assert reshape_dot(A=scalar, X=scalar, Y=scalar) is True
    assert reshape_dot(A=scalar, X=vec[2], Y=vec[2]) is False
    with pytest.raises(BuildError, match="shape mismatch"):
        reshape_dot(A=scalar, X=vec[3], Y=vec[1])

    # elif X.shape == ():
    assert reshape_dot(A=vec[1], X=scalar, Y=vec[1]) is True
    assert reshape_dot(A=vec[2], X=scalar, Y=vec[2]) is False
    with pytest.raises(BuildError, match="shape mismatch"):
        reshape_dot(A=vec[2], X=scalar, Y=vec[1])

    # elif X.ndim == 1:
    assert reshape_dot(A=vec[0], X=vec[0], Y=vec[0]) is False
    assert reshape_dot(A=vec[1], X=vec[1], Y=vec[1]) is True
    assert reshape_dot(A=vec[1], X=vec[1], Y=scalar) is True
    assert reshape_dot(A=vec[2], X=vec[2], Y=vec[2]) is False
    assert reshape_dot(A=mat23, X=vec[3], Y=vec[2]) is False
    with pytest.raises(BuildError, match="shape mismatch"):
        reshape_dot(A=mat23, X=vec[2], Y=vec[2])
    with pytest.raises(BuildError, match="shape mismatch"):
        reshape_dot(A=mat23, X=vec[3], Y=vec[1])

    # else:
    assert reshape_dot(A=mat11, X=mat11, Y=mat11) is True
    assert reshape_dot(A=mat23, X=mat33, Y=mat23) is False
    with pytest.raises(BuildError, match="shape mismatch"):
        reshape_dot(A=mat11, X=mat23, Y=mat23)
    with pytest.raises(BuildError, match="shape mismatch"):
        reshape_dot(A=mat23, X=mat33, Y=mat33)
    with pytest.raises(BuildError, match="shape mismatch"):
        reshape_dot(A=mat23, X=mat33, Y=vec[2])


def test_dotinc_op(rng):
    argnames = ["A", "X", "Y"]
    args = {
        "A": Signal(np.ones((3, 3)), name="Av"),
        "X": Signal(np.ones((3,)), name="Xv"),
        "Y": Signal(np.ones((3,)), name="Yv"),
    }
    _, sim = _test_operator_arg_attributes(DotInc, argnames, args=args)
    assert (
        str(sim)
        == "DotInc{Signal(name=Av, shape=(3, 3)), Signal(name=Xv, shape=(3,)) -> "
        "Signal(name=Yv, shape=(3,))}"
    )

    A = Signal(np.ones((2, 2)))
    X = Signal(np.ones((2, 2)))
    Y = Signal(np.ones(2))
    with pytest.raises(BuildError, match="X must be a column vector"):
        DotInc(A, X, Y)

    X = Signal(np.ones(2))
    Y = Signal(np.ones((2, 2)))
    with pytest.raises(BuildError, match="Y must be a column vector"):
        DotInc(A, X, Y)


def test_sparsedotinc_op():
    A = Signal(np.ones(2))
    X = Signal(np.ones(2))
    Y = Signal(np.ones(2))

    with pytest.raises(BuildError, match="must be a sparse Signal"):
        SparseDotInc(A, X, Y)


def test_simpyfunc_op():
    def my_func(t, x):
        return x

    argnames = ["output", "t", "x", "fn"]
    args = {"output": "outv", "t": "tv", "x": "xv", "fn": my_func}
    non_signals = ["fn"]
    _, sim = _test_operator_arg_attributes(
        SimPyFunc, argnames, args=args, non_signals=non_signals
    )
    assert str(sim) == "SimPyFunc{xv -> outv, fn='my_func'}"



import numpy as np
import pytest

from nengo.spa.pointer import SemanticPointer


def test_init(allclose):
    a = SemanticPointer([1, 2, 3, 4])
    assert len(a) == 4

    a = SemanticPointer([1, 2, 3, 4, 5])
    assert len(a) == 5

    a = SemanticPointer(list(range(100)))
    assert len(a) == 100

    a = SemanticPointer(27)
    assert len(a) == 27
    assert allclose(a.length(), 1)

    with pytest.raises(Exception):
        a = SemanticPointer(np.zeros(2, 2))

    with pytest.raises(Exception):
        a = SemanticPointer(-1)
    with pytest.raises(Exception):
        a = SemanticPointer(0)
    with pytest.raises(Exception):
        a = SemanticPointer(1.7)
    with pytest.raises(Exception):
        a = SemanticPointer(None)
    with pytest.raises(Exception):
        a = SemanticPointer(int)


def test_length(allclose):
    a = SemanticPointer([1, 1])
    assert allclose(a.length(), np.sqrt(2))
    a = SemanticPointer(10) * 1.2
    assert allclose(a.length(), 1.2)


def test_normalize(allclose):
    a = SemanticPointer([1, 1])
    a.normalize()
    assert allclose(a.length(), 1)


def test_str():
    a = SemanticPointer([1, 1])
    assert str(a) == str(np.array([1.0, 1.0]))


def test_randomize(rng, allclose):
    a = SemanticPointer(100, rng=rng)
    std = np.std(a.v)
    assert allclose(std, 1.0 / np.sqrt(len(a)), rtol=0.02)

    a = SemanticPointer(100)
    a.randomize(rng=rng)
    assert len(a) == 100
    std = np.std(a.v)
    assert allclose(std, 1.0 / np.sqrt(len(a)), rtol=0.02)

    a = SemanticPointer(5)
    a.randomize(N=100, rng=rng)
    assert len(a) == 100
    std = np.std(a.v)
    assert allclose(std, 1.0 / np.sqrt(len(a)), rtol=0.02)


def test_make_unitary(rng, allclose):
    a = SemanticPointer(100, rng=rng)
    a.make_unitary()
    assert allclose(1, a.length())
    assert allclose(1, (a * a).length())
    assert allclose(1, (a * a * a).length())


def test_add_sub(allclose):
    a = SemanticPointer(10)
    b = SemanticPointer(10)
    c = a.copy()
    d = b.copy()

    c += b
    d -= -a

    assert allclose((a + b).v, a.v + b.v)
    assert allclose((a + b).v, c.v)
    assert allclose((a + b).v, d.v)
    assert allclose((a + b).v, (a - (-b)).v)


def test_convolution(rng, allclose):
    a = SemanticPointer(64, rng=rng)
    b = SemanticPointer(64, rng=rng)
    identity = SemanticPointer(np.eye(64)[0])

    c = a.copy()
    c *= b

    answer = np.fft.ifft(np.fft.fft(a.v) * np.fft.fft(b.v)).real

    assert allclose((a * b).v, answer)
    assert allclose(a.convolve(b).v, answer)
    assert allclose(c.v, answer)
    assert allclose((a * identity).v, a.v)
    assert (a * b * ~b).compare(a) > 0.65


def test_multiply(allclose):
    a = SemanticPointer(50)

    assert allclose((a * 5).v, a.v * 5)
    assert allclose((5 * a).v, a.v * 5)
    assert allclose((a * 5.7).v, a.v * 5.7)
    assert allclose((5.7 * a).v, a.v * 5.7)
    assert allclose((0 * a).v, np.zeros(50))
    assert allclose((1 * a).v, a.v)

    with pytest.raises(Exception):
        a * None
    with pytest.raises(Exception):
        a * "string"


def test_compare(rng, allclose):
    a = SemanticPointer(50, rng=rng) * 10
    b = SemanticPointer(50, rng=rng) * 0.1

    assert a.compare(a) > 0.99
    assert a.compare(b) < 0.2
    assert allclose(a.compare(b), a.dot(b) / (a.length() * b.length()))


def test_dot(rng, allclose):
    a = SemanticPointer(50, rng=rng) * 1.1
    b = SemanticPointer(50, rng=rng) * (-1.5)
    assert allclose(a.dot(b), np.dot(a.v, b.v))


def test_distance(rng):
    a = SemanticPointer(50, rng=rng)
    b = SemanticPointer(50, rng=rng)
    assert a.distance(a) < 1e-5
    assert a.distance(b) > 0.9


def test_invert(allclose):
    a = SemanticPointer(50)
    assert a.v[0] == (~a).v[0]
    assert a.v[49] == (~a).v[1]
    assert allclose(a.v[1:], (~a).v[:0:-1])


def test_len():
    a = SemanticPointer(5)
    assert len(a) == 5

    a = SemanticPointer(list(range(10)))
    assert len(a) == 10


def test_copy(allclose):
    a = SemanticPointer(5)
    b = a.copy()
    assert a is not b
    assert a.v is not b.v
    assert allclose(a.v, b.v)


def test_mse(allclose):
    a = SemanticPointer(50)
    b = SemanticPointer(50)

    assert allclose(((a - b).length() ** 2) / 50, a.mse(b))


def test_conv_matrix(allclose):
    a = SemanticPointer(50)
    b = SemanticPointer(50)

    m = b.get_convolution_matrix()

    assert allclose((a * b).v, np.dot(m, a.v))



import warnings

import numpy as np

from nengo.base import NengoObject, NengoObjectParam, ObjView
from nengo.dists import DistOrArrayParam, Distribution
from nengo.ensemble import Ensemble, Neurons
from nengo.exceptions import ValidationError
from nengo.learning_rules import LearningRuleType, LearningRuleTypeParam
from nengo.neurons import Direct
from nengo.node import Node
from nengo.params import BoolParam, Default, FunctionInfo, Parameter
from nengo.rc import rc
from nengo.solvers import LstsqL2, SolverParam
from nengo.synapses import Lowpass, SynapseParam
from nengo.transforms import Dense, NoTransform
from nengo.utils.functions import function_name
from nengo.utils.numpy import is_array_like, is_iterable
from nengo.utils.stdlib import checked_call


class PrePostParam(NengoObjectParam):
    def coerce(self, conn, nengo_obj):  # pylint: disable=arguments-renamed
        if isinstance(nengo_obj, Connection):
            raise ValidationError(
                "Cannot connect to or from connections. "
                "Did you mean to connect to the connection's learning rule?",
                attr=self.name,
                obj=conn,
            )
        return super().coerce(conn, nengo_obj)


class ConnectionLearningRuleTypeParam(LearningRuleTypeParam):
    """Connection-specific validation for learning rules."""

    coerce_defaults = False

    def check_pre(self, conn, rule):
        pre = conn.pre_obj
        if rule.modifies in ("decoders", "weights"):
            # pre object must be neural
            if not isinstance(pre, (Ensemble, Neurons)):
                raise ValidationError(
                    "'pre' must be of type 'Ensemble' or 'Neurons' for "
                    f"learning rule '{rule}' (got type '{type(pre).__name__}')",
                    attr=self.name,
                    obj=conn,
                )
            if isinstance(pre, Ensemble) and isinstance(pre.neuron_type, Direct):
                raise ValidationError(
                    "'pre' cannot have neuron type 'Direct'. Connections from "
                    "'Direct' ensembles do not have decoders or weights.",
                    attr=self.name,
                    obj=conn,
                )

    def check_post(self, conn, rule):
        if rule.modifies == "encoders":
            if not isinstance(conn.post_obj, Ensemble):
                raise ValidationError(
                    "'post' must be of type 'Ensemble' (got "
                    f"'{type(conn.post_obj).__name__}') for learning rule '{rule}'",
                    attr=self.name,
                    obj=conn,
                )
            if conn._to_neurons:
                raise ValidationError(
                    "Cannot use an encoder learning rule with a direct neural "
                    "connection (since encoders are not used in that case).",
                    attr=self.name,
                    obj=conn,
                )
        else:
            if not isinstance(conn.post_obj, (Ensemble, Neurons, Node)):
                raise ValidationError(
                    "'post' must be of type 'Ensemble', 'Neurons' or 'Node' (got "
                    f"'{type(conn.post_obj).__name__}') for learning rule '{rule}'",
                    attr=self.name,
                    obj=conn,
                )

    def check_rule(self, conn, rule):  # pylint: disable=arguments-renamed
        super().check_rule(conn, rule)
        self.check_pre(conn, rule)
        self.check_post(conn, rule)
        self.check_transform(conn, rule)

    def check_transform(self, conn, rule):
        if not conn.has_weights and rule.modifies in ("weights", "decoders"):
            raise ValidationError(
                f"Learning rule '{rule}' cannot be applied to a connection that does "
                "not have weights (transform=None)",
                attr=self.name,
                obj=conn,
            )

        if rule.modifies == "weights":
            # If the rule modifies 'weights', then it must have full weights
            if not conn._to_neurons:
                raise ValidationError(
                    f"Learning rule '{rule}' can only be applied on connections to "
                    "neurons. Try setting `solver.weights` to True or "
                    "connecting between two Neurons objects.",
                    attr=self.name,
                    obj=conn,
                )

            # transform matrix must be 2D
            pre = conn.pre_obj
            pre_size = pre.n_neurons if isinstance(pre, Ensemble) else conn.pre.size_out
            post_size = conn.post.size_in
            if not conn.solver.weights and conn.transform.shape != (
                post_size,
                pre_size,
            ):
                raise ValidationError(
                    "Transform must be 2D array with shape post_neurons x "
                    f"pre_neurons ({pre_size}, {post_size})",
                    attr=self.name,
                    obj=conn,
                )


class ConnectionSolverParam(SolverParam):
    """Connection-specific validation for decoder solvers."""

    coerce_defaults = False

    def coerce(self, conn, solver):  # pylint: disable=arguments-renamed
        solver = super().coerce(conn, solver)
        if solver is not None:
            # it's true that setting the solver on any connection without a pre Ensemble
            # has no effect, but we only warn when ``weights=True`` because in this case
            # we can be sure that it's not the default solver
            if solver.weights and not isinstance(conn.pre_obj, Ensemble):
                warnings.warn(
                    "For connections from objects other than Ensembles, setting the "
                    "solver has no effect"
                )

            if solver.weights and not isinstance(conn.post_obj, Ensemble):
                warnings.warn(
                    "For connections to objects other than Ensembles, setting "
                    "`weights=True` on a solver has no effect"
                )

        return solver


class EvalPointsParam(DistOrArrayParam):
    coerce_defaults = False

    def coerce(self, conn, distorarray):  # pylint: disable=arguments-renamed
        """Eval points are only valid when pre is an ensemble."""
        if distorarray is not None and not isinstance(conn.pre, Ensemble):
            msg = (
                "eval_points are only valid on connections from ensembles "
                f"(got type '{type(conn.pre).__name__}')"
            )
            raise ValidationError(msg, attr=self.name, obj=conn)
        return super().coerce(conn, distorarray)


class ConnectionFunctionParam(Parameter):
    """Connection-specific validation for functions."""

    coerce_defaults = False

    def check_array(self, conn, ndarray):
        if not isinstance(conn.eval_points, np.ndarray):
            raise ValidationError(
                "In order to set 'function' to specific points, 'eval_points' "
                "must also be set to specific points.",
                attr=self.name,
                obj=conn,
            )

        if ndarray.ndim != 2:
            raise ValidationError(
                f"array must be 2D (got {ndarray.ndim}D)", attr=self.name, obj=conn
            )

        if ndarray.shape[0] != conn.eval_points.shape[0]:
            raise ValidationError(
                "Number of evaluation points must match number of function points "
                f"({ndarray.shape[0]} != {conn.eval_points.shape[0]})",
                attr=self.name,
                obj=conn,
            )

    def check_function_can_be_applied(self, conn, function_info):
        function, _ = function_info
        type_pre = type(conn.pre_obj).__name__

        if function is not None:
            if not isinstance(conn.pre_obj, (Node, Ensemble)):
                raise ValidationError(
                    "function can only be set for connections from an Ensemble"
                    f" or Node (got type '{type_pre}')",
                    attr=self.name,
                    obj=conn,
                )

            if isinstance(conn.pre_obj, Node) and conn.pre_obj.output is None:
                raise ValidationError(
                    "Cannot apply functions to passthrough nodes",
                    attr=self.name,
                    obj=conn,
                )

    def coerce(self, conn, function):  # pylint: disable=arguments-renamed
        function = super().coerce(conn, function)

        if function is None:
            function_info = FunctionInfo(function=None, size=None)
        elif isinstance(function, FunctionInfo):
            function_info = function
        elif is_array_like(function):
            array = np.array(function, copy=False, dtype=rc.float_dtype)
            self.check_array(conn, array)
            function_info = FunctionInfo(function=array, size=array.shape[1])
        elif callable(function):
            function_info = FunctionInfo(
                function=function, size=self.determine_size(conn, function)
            )
            # TODO: necessary?
            super().coerce(conn, function_info)
        else:
            raise ValidationError(
                f"Invalid connection function type '{type(function).__name__}' "
                "(must be callable or array-like)",
                attr=self.name,
                obj=conn,
            )

        self.check_function_can_be_applied(conn, function_info)

        return function_info

    def determine_size(self, instance, function):
        args = self.function_args(instance, function)
        value, invoked = checked_call(function, *args)
        if not invoked:
            raise ValidationError(
                f"function '{function}' must accept a single np.array argument",
                attr=self.name,
                obj=instance,
            )
        return np.asarray(value).size

    def function_args(self, conn, function):
        x = (
            conn.eval_points[0]
            if is_iterable(conn.eval_points)
            else np.zeros(conn.size_in)
        )
        return (x,)


class ConnectionTransformParam(Parameter):
    """Connection-specific validation for transforms."""

    coerce_defaults = False

    def coerce(self, conn, transform):  # pylint: disable=arguments-renamed
        if transform is None:
            transform = NoTransform(conn.size_mid)
        elif is_array_like(transform) or isinstance(transform, Distribution):
            transform = Dense((conn.size_out, conn.size_mid), transform)

        if transform.size_in != conn.size_mid:
            if isinstance(transform, Dense) and (
                transform.shape[0] == transform.shape[1]
            ):
                # we provide a different error message in this case;
                # the transform is not changing the dimensionality of the
                # signal, so the blame most likely lies with the function
                raise ValidationError(
                    "Function output size is incorrect; should return a "
                    f"vector of size {conn.size_mid}",
                    attr=self.name,
                    obj=conn,
                )
            else:
                raise ValidationError(
                    f"Transform input size ({transform.size_in}) not equal to "
                    f"'{type(conn.pre_obj).__name__}' output size ({conn.size_mid})",
                    attr=self.name,
                    obj=conn,
                )

        if transform.size_out != conn.size_out:
            raise ValidationError(
                f"Transform output size ({transform.size_out}) not equal to connection "
                f"output size ({conn.size_out})",
                attr=self.name,
                obj=conn,
            )

        # we don't support repeated indices on 2D transforms because it makes
        # the matrix multiplication more complicated (we'd need to expand
        # the weight matrix for the duplicated rows/columns). it could be done
        # if there were a demand at some point.
        if isinstance(transform, Dense) and len(transform.init_shape) == 2:

            def repeated_inds(x):
                return not isinstance(x, slice) and np.unique(x).size != len(x)

            if repeated_inds(conn.pre_slice):
                raise ValidationError(
                    "Input object selection has repeated indices",
                    attr=self.name,
                    obj=conn,
                )
            if repeated_inds(conn.post_slice):
                raise ValidationError(
                    "Output object selection has repeated indices",
                    attr=self.name,
                    obj=conn,
                )

        return super().coerce(conn, transform)


class Connection(NengoObject):
    """Connects two objects together.

    The connection between the two object is unidirectional,
    transmitting information from the first argument, ``pre``,
    to the second argument, ``post``.

    Almost any Nengo object can act as the pre or post side of a connection.
    Additionally, you can use Python slice syntax to access only some of the
    dimensions of the pre or post object.

    For example, if ``node`` has ``size_out=2`` and ``ensemble`` has
    ``size_in=1``:

    .. testcode::

       with nengo.Network() as net:
           node = nengo.Node(np.zeros(2))
           ensemble = nengo.Ensemble(10, 1)

    We could not create the following connection:

    .. testcode::

       with net:
           nengo.Connection(node, ensemble)

    .. testoutput::
       :hide:

       Traceback (most recent call last):
       ...
       nengo.exceptions.ValidationError: init: Shape of initial value () does not \
       match expected shape (1, 2)

    But, we could create either of these two connections:

    .. testcode::

       with net:
           nengo.Connection(node[0], ensemble)
           nengo.Connection(node[1], ensemble)

    Parameters
    ----------
    pre : Ensemble or Neurons or Node
        The source Nengo object for the connection.
    post : Ensemble or Neurons or Node or LearningRule
        The destination object for the connection.
    synapse : Synapse or None, optional
        Synapse model to use for filtering (see `~nengo.synapses.Synapse`).
        If *None*, no synapse will be used and information will be transmitted
        without any delay (if supported by the backend---some backends may
        introduce a single time step delay).

        Note that at least one connection must have a synapse that is not
        *None* if components are connected in a cycle. Furthermore, a synaptic
        filter with a zero time constant is different from a *None* synapse
        as a synaptic filter will always add a delay of at least one time step.
    function : callable or (n_eval_points, size_mid) array_like, optional
        Function to compute across the connection. Note that ``pre`` must be
        an ensemble to apply a function across the connection.
        If an array is passed, the function is implicitly defined by the
        points in the array and the provided ``eval_points``, which have a
        one-to-one correspondence.
    transform : (size_out, size_mid) array_like, optional
        Linear transform mapping the pre output to the post input.
        This transform is in terms of the sliced size; if either pre
        or post is a slice, the transform must be shaped according to
        the sliced dimensionality. Additionally, the function is applied
        before the transform, so if a function is computed across the
        connection, the transform must be of shape ``(size_out, size_mid)``.
    solver : Solver, optional
        Solver instance to compute decoders or weights
        (see `~nengo.solvers.Solver`). If ``solver.weights`` is True, a full
        connection weight matrix is computed instead of decoders.
    learning_rule_type : LearningRuleType or iterable of LearningRuleType, optional
        Modifies the decoders or connection weights during simulation.
    eval_points : (n_eval_points, size_in) array_like or int, optional
        Points at which to evaluate ``function`` when computing decoders,
        spanning the interval (-pre.radius, pre.radius) in each dimension.
        If None, will use the eval_points associated with ``pre``.
    scale_eval_points : bool, optional
        Indicates whether the evaluation points should be scaled
        by the radius of the pre Ensemble.
    label : str, optional
        A descriptive label for the connection.
    seed : int, optional
        The seed used for random number generation.

    Attributes
    ----------
    function : callable
        The given function.
    function_size : int
        The output dimensionality of the given function. If no function is
        specified, function_size will be 0.
    label : str
        A human-readable connection label for debugging and visualization.
        If not overridden, incorporates the labels of the pre and post objects.
    learning_rule_type : instance or list or dict of LearningRuleType, optional
        The learning rule types.
    post : Ensemble or Neurons or Node or Probe or ObjView
        The given post object.
    post_obj : Ensemble or Neurons or Node or Probe
        The underlying post object, even if ``post`` is an ``ObjView``.
    post_slice : slice or list or None
        The slice associated with ``post`` if it is an ObjView, or None.
    pre : Ensemble or Neurons or Node or ObjView
        The given pre object.
    pre_obj : Ensemble or Neurons or Node
        The underlying pre object, even if ``post`` is an ``ObjView``.
    pre_slice : slice or list or None
        The slice associated with ``pre`` if it is an ObjView, or None.
    seed : int
        The seed used for random number generation.
    solver : Solver
        The Solver instance that will be used to compute decoders or weights
        (see ``nengo.solvers``).
    synapse : Synapse
        The Synapse model used for filtering across the connection
        (see ``nengo.synapses``).
    transform : (size_out, size_mid) array_like
        Linear transform mapping the pre function output to the post input.

    Properties
    ----------
    learning_rule : LearningRule or iterable of LearningRule
        Connectable learning rule object(s) associated with this connection.
    size_in : int
        The number of output dimensions of the pre object.
        Also the input size of the function, if one is specified.
    size_mid : int
        The number of output dimensions of the function, if specified.
        If the function is not specified, then ``size_in == size_mid``.
    size_out : int
        The number of input dimensions of the post object.
        Also the number of output dimensions of the transform.
    """

    probeable = ("output", "input", "weights")

    pre = PrePostParam("pre", nonzero_size_out=True)
    post = PrePostParam("post", nonzero_size_in=True)
    synapse = SynapseParam("synapse", default=Lowpass(tau=0.005))
    function_info = ConnectionFunctionParam("function", default=None, optional=True)
    transform = ConnectionTransformParam("transform", default=None, optional=True)
    solver = ConnectionSolverParam("solver", default=LstsqL2())
    learning_rule_type = ConnectionLearningRuleTypeParam(
        "learning_rule_type", default=None, optional=True
    )
    eval_points = EvalPointsParam(
        "eval_points", default=None, optional=True, sample_shape=("*", "size_in")
    )
    scale_eval_points = BoolParam("scale_eval_points", default=True)

    _param_init_order = [
        "pre",
        "post",
        "synapse",
        "eval_points",
        "function_info",
        "transform",
        "solver",
        "learning_rule_type",
    ]

    def __init__(
        self,
        pre,
        post,
        synapse=Default,
        function=Default,
        transform=Default,
        solver=Default,
        learning_rule_type=Default,
        eval_points=Default,
        scale_eval_points=Default,
        label=Default,
        seed=Default,
    ):
        super().__init__(label=label, seed=seed)

        self.pre = pre
        self.post = post

        self.synapse = synapse
        self.eval_points = eval_points  # Must be set before function
        self.scale_eval_points = scale_eval_points
        self.function_info = function
        self.transform = transform  # Must be set after function
        self.solver = solver  # Must be set before learning rule
        self.learning_rule_type = learning_rule_type  # set after transform

    def __str__(self):
        return self._str(include_id=False)

    def __repr__(self):
        return self._str(include_id=True)

    def _str(self, include_id):
        desc = "<Connection "
        if include_id:
            desc += f"at 0x{id(self):x} "

        if self.label is None:
            func_txt = (
                ""
                if self.function is None
                else f" computing '{function_name(self.function)}'"
            )
            desc += f"from {self.pre} to {self.post}{func_txt}"
        else:
            desc += self.label

        desc += ">"

        return desc

    @property
    def function(self):
        return self.function_info.function

    @function.setter
    def function(self, function):
        self.function_info = function

    @property
    def has_weights(self):
        return not isinstance(self.transform, NoTransform) or (
            isinstance(self.pre_obj, Ensemble)
            and not isinstance(self.pre_obj.neuron_type, Direct)
        )

    @property
    def is_decoded(self):
        warnings.warn(
            "is_decoded is deprecated; directly check the pre/post objects for the "
            "properties of interest instead",
            DeprecationWarning,
        )
        return not (
            self.solver.weights
            or (
                isinstance(self.pre_obj, Neurons) and isinstance(self.post_obj, Neurons)
            )
        )

    @property
    def _to_neurons(self):
        return isinstance(self.post_obj, Neurons) or (
            isinstance(self.pre_obj, Ensemble)
            and isinstance(self.post_obj, Ensemble)
            and self.solver.weights
        )

    @property
    def learning_rule(self):
        """(LearningRule or iterable) Connectable learning rule object(s)."""
        if self.learning_rule_type is None:
            return None

        types = self.learning_rule_type
        if isinstance(types, dict):
            learning_rule = type(types)()  # dict of same type
            for k, v in types.items():
                learning_rule[k] = LearningRule(self, v)
        elif is_iterable(types):
            learning_rule = [LearningRule(self, v) for v in types]
        elif isinstance(types, LearningRuleType):
            learning_rule = LearningRule(self, types)
        else:
            assert False, "Validation should catch this"

        return learning_rule

    @property
    def post_obj(self):
        return self.post.obj if isinstance(self.post, ObjView) else self.post

    @property
    def post_slice(self):
        return self.post.slice if isinstance(self.post, ObjView) else slice(None)

    @property
    def pre_obj(self):
        return self.pre.obj if isinstance(self.pre, ObjView) else self.pre

    @property
    def pre_slice(self):
        return self.pre.slice if isinstance(self.pre, ObjView) else slice(None)

    @property
    def size_in(self):
        """(int) The number of output dimensions of the pre object.

        Also the input size of the function, if one is specified.
        """
        return self.pre.size_out

    @property
    def size_mid(self):
        """(int) The number of output dimensions of the function, if specified.

        If the function is not specified, then ``size_in == size_mid``.
        """
        size = self.function_info.size
        return self.size_in if size is None else size

    @property
    def size_out(self):
        """(int) The number of input dimensions of the post object.

        Also the number of output dimensions of the transform.
        """
        return self.post.size_in


class LearningRule:
    """An interface for making connections to a learning rule.

    Connections to a learning rule are to allow elements of the network to
    affect the learning rule. For example, learning rules that use error
    information can obtain that information through a connection.

    Learning rule objects should only ever be accessed through the
    ``learning_rule`` attribute of a connection.
    """

    def __init__(self, connection, learning_rule_type):
        self._connection = connection
        self.learning_rule_type = learning_rule_type

    def __repr__(self):
        return (
            f"<LearningRule at 0x{id(self):x} modifying {self.connection!r} "
            f"with type {self.learning_rule_type!r}>"
        )

    def __str__(self):
        return (
            f"<LearningRule modifying {self.connection} "
            f"with type {self.learning_rule_type}>"
        )

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self._connection is other._connection
            and self.learning_rule_type == other.learning_rule_type
        )

    def __hash__(self):
        # +1 to avoid collision with ensemble
        return hash(self._connection) + hash(self.learning_rule_type) + 1

    def __getitem__(self, key):
        return ObjView(self, key)

    @property
    def connection(self):
        """(Connection) The connection modified by the learning rule."""
        return self._connection

    @property
    def modifies(self):
        """(str) The variable modified by the learning rule."""
        return self.learning_rule_type.modifies

    @property
    def probeable(self):
        """(tuple) Signals that can be probed in the learning rule."""
        return self.learning_rule_type.probeable

    @property
    def size_in(self):
        conn = self.connection
        size_in = self.learning_rule_type.size_in
        if size_in == "pre":
            return conn.size_in
        elif size_in == "mid":
            return conn.size_mid
        elif size_in == "post":
            return conn.size_out
        elif size_in == "pre_state":
            return (
                conn.pre_obj.ensemble.size_out
                if isinstance(conn.pre_obj, Neurons)
                else conn.size_in
            )
        elif size_in == "post_state":
            return (
                conn.post_obj.ensemble.size_in
                if isinstance(conn.post_obj, Neurons)
                else conn.size_out
            )
        else:
            return size_in  # should be an integer

    @property
    def size_out(self):
        """(int) Cannot connect from learning rules, so always 0."""
        return 0  # since a learning rule can't connect to anything
        # TODO: allow probing individual learning rules



from nengo._vendor import portalocker
from nengo.exceptions import TimeoutError


class FileLock:
    """Lock access to a file (for multithreading)."""

    def __init__(self, filename, timeout=10.0, poll=0.1):
        self.filename = filename
        self.timeout = timeout
        self.poll = poll
        self._lock = portalocker.Lock(
            self.filename, timeout=timeout, check_interval=poll, fail_when_locked=True
        )
        self._acquired = False

    def acquire(self):
        try:
            self._lock.acquire()
            self._acquired = True
        except (portalocker.AlreadyLocked, portalocker.LockException) as e:
            raise TimeoutError(f"Could not acquire lock '{self.filename}'.") from e

    def release(self):
        self._lock.release()
        self._acquired = False

    @property
    def acquired(self):
        return self._acquired

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    def __del__(self):
        self.release()



import numpy as np
import pytest

from nengo.utils.filter_design import (
    _none_to_empty_2d,
    _restore,
    _shape_or_none,
    abcd_normalize,
    cont2discrete,
    normalize,
    ss2tf,
    ss2zpk,
    tf2ss,
    tf2zpk,
    zpk2ss,
    zpk2tf,
)


@pytest.mark.parametrize("dt", [1e-3, 1e-2, 1e-1])
def test_cont2discrete_zoh(dt, allclose):
    """test the function cont2discrete with zero-order hold"""
    taus = np.logspace(-np.log10(dt) - 1, np.log10(dt) + 3, 30)

    # test with lowpass filter, using analytic solution
    for tau in taus:
        num, den = [1], [tau, 1]
        d = -np.expm1(-dt / tau)
        num0, den0 = [0, d], [1, d - 1]
        num1, den1, _ = cont2discrete((num, den), dt)
        assert allclose(num0, num1)
        assert allclose(den0, den1)

    # test with alpha filter, using analytic solution
    for tau in taus:
        num, den = [1], [tau ** 2, 2 * tau, 1]
        a = dt / tau
        ea = np.exp(-a)
        num0 = [0, -a * ea + (1 - ea), ea * (a + ea - 1)]
        den0 = [1, -2 * ea, ea ** 2]
        num1, den1, _ = cont2discrete((num, den), dt)
        assert allclose(num0, num1)
        assert allclose(den0, den1)

    # test integrative filter, using analytic solution
    num, den = [1], [1, 0]
    num0, den0 = [0, dt], [1, -1]
    num1, den1, _ = cont2discrete((num, den), dt)
    assert allclose(num0, num1)
    assert allclose(den0, den1)


def test_cont2discrete_other_methods():
    dt = 1e-3

    # test with len(sys) == 3
    assert (
        repr(cont2discrete(([1], [1], [1]), dt))
        == "(array([1.0010005]), array([1.0010005]), 1.0, 0.001)"
    )

    # test with len(sys) == 5
    with pytest.raises(ValueError):
        cont2discrete(([1], [1], [1], [1], [1]), dt)

    # test method gbt and alpha None
    with pytest.raises(ValueError):
        cont2discrete(([1], [1], [1]), dt, method="gbt", alpha=None)

    # test method gbt and alpha invalid
    with pytest.raises(ValueError):
        cont2discrete(([1], [1], [1]), dt, method="gbt", alpha=2)

    # test method bilinear
    assert (
        repr(cont2discrete(([1], [1], [1]), dt, method="bilinear"))
        == "(array([1.0010005]), array([1.0010005]), 1.0, 0.001)"
    )

    # test method backward_diff
    assert (
        repr(cont2discrete(([1], [1], [1]), dt, method="backward_diff"))
        == "(array([1.001001]), array([1.001001]), 1.0, 0.001)"
    )

    # test bad method
    with pytest.raises(ValueError):
        cont2discrete(([1], [1], [1]), dt, method="not_a_method")


def test_tf2zpk():
    """test the tf2zpk function"""
    (a, b, c) = tf2zpk(1, 2)
    assert a.size == 0
    assert b.size == 0
    assert c == 0.5


def test_zpk2tf():
    """test the zpk2tf function"""
    (b, a) = zpk2tf([[1], [2]], [3], 3.0)
    z = [[1], [2]]
    z = np.atleast_1d(z)
    assert len(z.shape) > 1
    assert a.all() == 1
    assert b.all() == 1

    (b, a) = zpk2tf([0], [1], 3.0)
    z = [0]
    z = np.atleast_1d(z)
    assert len(z.shape) <= 1
    assert a.all() == 1
    assert b.all() == 0


@pytest.mark.filterwarnings("ignore:Badly conditioned filter coefficients")
def test_normalize():
    """test the normalize function, and exceptions"""
    with pytest.raises(ValueError):
        a = [[1], [2]]
        b = [1]
        normalize(b, a)
    with pytest.raises(ValueError):
        a = [1]
        b = [[[1]]]
        normalize(b, a)
    with pytest.warns(Warning):
        a = [1]
        b = [0]
        normalize(b, a)
    a = [1]
    b = [0, 0, 0]
    normalize(b, a)


def test_tf2ss():
    """test the tf2ss function, and exceptions"""
    with pytest.raises(ValueError):
        num = [1, 2]
        den = [1]
        tf2ss(num, den)


def test_none_to_empty_2d():
    """test the _none_to_empty_2d function"""
    assert np.array_equal(np.zeros((0, 0)), _none_to_empty_2d(None))


def test_shape_or_none():
    """test the _shape_or_none function"""
    assert _shape_or_none(None) == (None,) * 2


def test_restore(allclose):
    """test the _restore function and errors"""

    class Test:
        shape = (0, 0)

    M = Test
    shape = (0, 0)
    assert allclose(_restore(M, shape), np.zeros(shape))
    with pytest.raises(ValueError):
        M = np.array([1, 2])
        shape = (1, 2)
        _restore(M, shape)


def test_abcd_normalize():
    """test the function abcd_normalize"""
    with pytest.raises(ValueError):
        abcd_normalize(None, None, None, None)


def test_ss2tf():
    """test the function ss2tf"""
    with pytest.raises(ValueError):
        ss2tf(None, None, None, None, 5)


def test_zpk2ss():
    """test the function zpk2ss"""

    predicted = (
        np.array([[3.0, -2.0], [1.0, 0.0]]),
        np.array([[1.0], [0.0]]),
        np.array([[3.0, -3.0]]),
        np.array([0.0]),
    )

    assert repr(zpk2ss([1], [1, 2], 3)) == repr(predicted)


def test_ss2zpk():
    """test the function ss2zpk"""
    predicted = (np.array([0.0]), np.array([1.0]), 1.0)
    assert repr(ss2zpk([1], [1], [1], [1])) == repr(predicted)



import pickle

import numpy as np
import pytest

import nengo
import nengo.utils.numpy as npext
from nengo.dists import Choice, Uniform, UniformHypersphere
from nengo.exceptions import BuildError, NengoWarning, ReadonlyError
from nengo.neurons import RegularSpiking
from nengo.processes import FilteredNoise, WhiteNoise
from nengo.utils.testing import signals_allclose


def test_missing_attribute():
    with nengo.Network():
        a = nengo.Ensemble(10, 1)

        with pytest.warns(SyntaxWarning):
            a.new_attribute = 9


@pytest.mark.parametrize("dimensions", [1, 200])
def test_encoders(Simulator, dimensions, seed, allclose, n_neurons=10, encoders=None):
    dtype = nengo.rc.float_dtype
    if encoders is None:
        encoders = np.random.normal(size=(n_neurons, dimensions))
        encoders = npext.array(encoders, min_dims=2, dtype=dtype)
        encoders /= npext.norm(encoders, axis=1, keepdims=True)

    model = nengo.Network(label="_test_encoders", seed=seed)
    with model:
        ens = nengo.Ensemble(
            n_neurons=n_neurons, dimensions=dimensions, encoders=encoders, label="A"
        )

    with Simulator(model) as sim:
        assert allclose(encoders, sim.data[ens].encoders)


def test_encoders_wrong_shape(Simulator, seed, allclose):
    dimensions = 3
    encoders = np.random.normal(size=dimensions)
    with pytest.raises(ValueError):
        test_encoders(
            Simulator, dimensions, seed=seed, encoders=encoders, allclose=allclose
        )


def test_encoders_negative_neurons(Simulator, seed, allclose):
    with pytest.raises(ValueError):
        test_encoders(Simulator, 1, seed=seed, n_neurons=-1, allclose=allclose)


def test_encoders_no_dimensions(Simulator, seed, allclose):
    with pytest.raises(ValueError):
        test_encoders(Simulator, 0, seed=seed, allclose=allclose)


@pytest.mark.filterwarnings("ignore:invalid value encountered in true_divide")
def test_encoder_all_zero(Simulator, seed, allclose):
    with pytest.raises(BuildError, match="zero-length encoders"):
        test_encoders(
            Simulator, 1, seed=seed, allclose=allclose, encoders=np.zeros((10, 1))
        )


def test_constant_scalar(Simulator, AnyNeuronType, plt, seed, allclose):
    """A Network that represents a constant value."""
    N = 30
    val = 0.5

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        input = nengo.Node(output=val, label="input")
        A = nengo.Ensemble(N, 1)
        nengo.Connection(input, A)
        in_p = nengo.Probe(input, "output")
        A_p = nengo.Probe(A, "decoded_output", synapse=0.05)

    with Simulator(m) as sim:
        sim.run(0.3)

    t = sim.trange()
    plt.plot(t, sim.data[in_p], label="Input")
    plt.plot(t, sim.data[A_p], label="Neuron approximation, pstc=0.05")
    plt.ylim(top=1.05 * val)
    plt.xlim(right=t[-1])
    plt.legend(loc=0)

    assert allclose(sim.data[in_p], val, atol=0.1, rtol=0.01)
    assert allclose(sim.data[A_p][-10:], val, atol=0.1, rtol=0.01)


def test_constant_vector(Simulator, AnyNeuronType, plt, seed, allclose):
    """A network that represents a constant 3D vector."""
    N = 30
    vals = [0.6, 0.1, -0.5]

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        input = nengo.Node(output=vals)
        A = nengo.Ensemble(N * len(vals), len(vals))
        nengo.Connection(input, A)
        in_p = nengo.Probe(input, "output")
        A_p = nengo.Probe(A, "decoded_output", synapse=0.05)

    with Simulator(m) as sim:
        sim.run(0.3)

    t = sim.trange()
    plt.plot(t, sim.data[in_p], label="Input")
    plt.plot(t, sim.data[A_p], label="Neuron approximation, pstc=0.05")
    plt.legend(loc="best", fontsize="small")
    plt.xlim(right=t[-1])

    assert allclose(sim.data[in_p][-10:], vals, atol=0.1, rtol=0.01)
    assert allclose(sim.data[A_p][-10:], vals, atol=0.1, rtol=0.01)


def test_scalar(Simulator, AnyNeuronType, plt, seed, allclose):
    """A network that represents sin(t)."""
    N = 50
    f = lambda t: np.sin(2 * np.pi * t)

    m = nengo.Network(label="test_scalar", seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        input = nengo.Node(output=f)
        A = nengo.Ensemble(N, 1, label="A")
        nengo.Connection(input, A)
        in_p = nengo.Probe(input, "output")
        A_p = nengo.Probe(A, "decoded_output", synapse=0.02)

    with Simulator(m) as sim:
        sim.run(1.0)
    t = sim.trange()
    target = f(t)

    assert signals_allclose(
        t, target, sim.data[in_p], rtol=1e-3, atol=1e-5, allclose=allclose
    )
    assert signals_allclose(
        t, target, sim.data[A_p], atol=0.1, delay=0.02, plt=plt, allclose=allclose
    )


def test_vector(Simulator, AnyNeuronType, plt, seed, allclose):
    """A network that represents sin(t), cos(t), cos(t)**2."""
    N = 100
    f = lambda t: [np.sin(6.3 * t), np.cos(6.3 * t), np.cos(6.3 * t) ** 2]

    m = nengo.Network(label="test_vector", seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        input = nengo.Node(output=f)
        A = nengo.Ensemble(N * 3, 3, radius=1.5)
        nengo.Connection(input, A)
        in_p = nengo.Probe(input)
        A_p = nengo.Probe(A, synapse=0.03)

    with Simulator(m) as sim:
        sim.run(1.0)
    t = sim.trange()
    target = np.vstack(f(t)).T

    assert signals_allclose(
        t, target, sim.data[in_p], rtol=1e-3, atol=1e-5, allclose=allclose
    )
    assert signals_allclose(
        t,
        target,
        sim.data[A_p],
        plt=plt,
        atol=0.1,
        delay=0.03,
        buf=0.1,
        allclose=allclose,
    )


def test_product(Simulator, AnyNeuronType, plt, seed):
    N = 80
    dt2 = 0.002
    scaling = -0.5
    f = lambda t: np.sin(2 * np.pi * t)

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = AnyNeuronType()
        sin = nengo.Node(output=f)
        cons = nengo.Node(output=scaling)
        factors = nengo.Ensemble(
            2 * N,
            dimensions=2,
            radius=1.5,
            encoders=Choice([[1, 1], [-1, 1], [1, -1], [-1, -1]]),
        )
        product = nengo.Ensemble(N, dimensions=1)
        nengo.Connection(sin, factors[0])
        nengo.Connection(cons, factors[1])
        nengo.Connection(factors, product, function=lambda x: x[0] * x[1], synapse=0.01)

        factors_p = nengo.Probe(factors, sample_every=dt2, synapse=0.01)
        product_p = nengo.Probe(product, sample_every=dt2, synapse=0.01)

    with Simulator(m) as sim:
        sim.run(1)
    t = sim.trange(sample_every=dt2)

    plt.subplot(211)
    plt.plot(t, sim.data[factors_p])
    plt.legend(["factor 1", "factor 2"], loc="best")
    plt.subplot(212)
    plt.plot(t, -0.5 * f(t), "k--")
    plt.plot(t, sim.data[product_p])
    plt.legend(["exact product", "neural product"], loc="best")

    assert npext.rms(sim.data[factors_p][:, 0] - f(t)) < 0.1
    assert npext.rms(sim.data[factors_p][20:, 1] - scaling) < 0.1
    assert npext.rms(sim.data[product_p][:, 0] - scaling * f(t)) < 0.1


@pytest.mark.parametrize("dims, points", [(1, 528), (2, 823), (3, 937)])
def test_eval_points_number(Simulator, dims, points, seed):
    model = nengo.Network(seed=seed)
    with model:
        A = nengo.Ensemble(5, dims, n_eval_points=points)

    with Simulator(model) as sim:
        assert sim.data[A].eval_points.shape == (points, dims)


def test_eval_points_number_warning(Simulator, seed, allclose):
    model = nengo.Network(seed=seed)
    with model:
        A = nengo.Ensemble(5, 1, n_eval_points=10, eval_points=[[0.1], [0.2]])

    with pytest.warns(UserWarning):
        # n_eval_points doesn't match actual passed eval_points, which warns
        with Simulator(model) as sim:
            pass

    assert allclose(sim.data[A].eval_points, [[0.1], [0.2]])


def test_gain_bias_warning(Simulator, seed):
    model = nengo.Network(seed=seed)
    with model:
        nengo.Ensemble(1, 1, gain=[1], bias=[1], intercepts=[0.5])
        nengo.Ensemble(1, 1, gain=[1], bias=[1], max_rates=[200])

    with pytest.warns(NengoWarning):
        # gain, bias override specified intercepts and max_rates, which warns
        with Simulator(model):
            pass


@pytest.mark.parametrize(
    "neurons, dims", [(10, 1), (392, 1), (2108, 1), (100, 2), (1290, 4), (20, 9)]
)
def test_eval_points_heuristic(Simulator, neurons, dims, seed):
    def heuristic(neurons, dims):
        return max(npext.clip(500 * dims, 750, 2500), 2 * neurons)

    model = nengo.Network(seed=seed)
    with model:
        A = nengo.Ensemble(neurons, dims)

    with Simulator(model) as sim:
        points = sim.data[A].eval_points
        assert points.shape == (heuristic(neurons, dims), dims)


@pytest.mark.parametrize("sample", [False, True])
@pytest.mark.parametrize("radius", [0.5, 1, 1.5])
def test_eval_points_scaling(Simulator, sample, radius, seed, rng):
    eval_points = UniformHypersphere()
    if sample:
        eval_points = eval_points.sample(500, 3, rng=rng)

    model = nengo.Network(seed=seed)
    with model:
        a = nengo.Ensemble(1, 3, eval_points=eval_points, radius=radius)

    with Simulator(model) as sim:
        dists = npext.norm(sim.data[a].eval_points, axis=1)
        assert np.all(dists <= radius)
        assert np.any(dists >= 0.9 * radius)


def test_len():
    """Make sure we can do len(ens) or len(ens.neurons)."""
    with nengo.Network():
        ens1 = nengo.Ensemble(10, dimensions=1)
        ens5 = nengo.Ensemble(100, dimensions=5)
    # Ensemble.__len__
    assert len(ens1) == 1
    assert len(ens5) == 5
    assert len(ens1[0]) == 1
    assert len(ens1[-1]) == 1
    assert len(ens5[:3]) == 3
    assert len(ens5[:100]) == 5

    # Neurons.__len__
    assert len(ens1.neurons) == 10
    assert len(ens5.neurons) == 100
    assert len(ens1.neurons[0]) == 1
    assert len(ens5.neurons[90:]) == 10
    assert len(ens5.neurons[90:150]) == 10


def test_invalid_indices():
    with nengo.Network():
        ens = nengo.Ensemble(10, dimensions=2)

    with pytest.raises(IndexError):
        assert ens[5]
    with pytest.raises(IndexError):
        assert ens[-5]
    with pytest.raises(IndexError):
        assert ens.neurons[20]


def test_invalid_rates(Simulator):
    with nengo.Network() as model:
        nengo.Ensemble(1, 1, max_rates=[200], neuron_type=nengo.LIF(tau_ref=0.01))

    with pytest.raises(ValueError):
        with Simulator(model):
            pass


def test_gain_bias(Simulator):
    N = 17
    D = 2

    gain = np.random.uniform(low=0.2, high=5, size=N)
    bias = np.random.uniform(low=1.2, high=2, size=N)

    model = nengo.Network()
    with model:
        a = nengo.Ensemble(N, D)
        a.gain = gain
        a.bias = bias

    with Simulator(model) as sim:
        assert np.array_equal(gain, sim.data[a].gain)
        assert np.array_equal(bias, sim.data[a].bias)


def test_noise_gen(Simulator, NonDirectNeuronType, seed, plt, allclose):
    """Ensure that setting Ensemble.noise generates noise."""
    with nengo.Network(seed=seed) as model:
        intercepts = -0.5
        neg_noise, pos_noise = -5, 5
        model.config[nengo.Ensemble].neuron_type = NonDirectNeuronType()
        model.config[nengo.Ensemble].encoders = Choice([[1]])
        model.config[nengo.Ensemble].intercepts = Choice([intercepts])
        pos = nengo.Ensemble(1, 1, noise=WhiteNoise(Uniform(0, pos_noise)))
        normal = nengo.Ensemble(1, 1)
        neg = nengo.Ensemble(1, 1, noise=WhiteNoise(Uniform(neg_noise, 0)))
        pos_p = nengo.Probe(pos.neurons, synapse=0.1)
        normal_p = nengo.Probe(normal.neurons, synapse=0.1)
        neg_p = nengo.Probe(neg.neurons, synapse=0.1)
    with Simulator(model) as sim:
        sim.run(0.06)

    t = sim.trange()
    plt.title(f"intercepts={intercepts}")
    plt.plot(t, sim.data[pos_p], c="b", label=f"noise={pos_noise}")
    plt.plot(t, sim.data[normal_p], c="k", label="no noise")
    plt.plot(t, sim.data[neg_p], c="r", label=f"noise={neg_noise}")
    plt.legend(loc="best")

    assert np.sum(sim.data[pos_p], axis=0) >= np.sum(sim.data[normal_p], axis=0)
    assert np.sum(sim.data[normal_p], axis=0) >= np.sum(sim.data[neg_p], axis=0)
    assert not allclose(
        sim.data[normal_p], sim.data[pos_p], record_rmse=False, print_fail=0
    )
    assert not allclose(
        sim.data[normal_p], sim.data[neg_p], record_rmse=False, print_fail=0
    )


def test_noise_copies_ok(Simulator, NonDirectNeuronType, seed, plt, allclose):
    """Make sure the same noise process works in multiple ensembles.

    We test this both with the default system and without.
    """

    process = FilteredNoise(synapse=nengo.Alpha(1.0), dist=Choice([[0.5]]))
    with nengo.Network(seed=seed) as model:
        if (
            NonDirectNeuronType.spiking
            or RegularSpiking in NonDirectNeuronType.__bases__
        ):
            neuron_type = NonDirectNeuronType(initial_state={"voltage": Choice([0])})
        else:
            neuron_type = NonDirectNeuronType()
        model.config[nengo.Ensemble].neuron_type = neuron_type
        model.config[nengo.Ensemble].encoders = Choice([[1]])
        model.config[nengo.Ensemble].gain = Choice([5])
        model.config[nengo.Ensemble].bias = Choice([2])
        model.config[nengo.Ensemble].noise = process
        const = nengo.Node(output=1)
        a = nengo.Ensemble(1, 1, noise=process)
        b = nengo.Ensemble(1, 1, noise=process)
        c = nengo.Ensemble(1, 1)  # defaults to noise=process
        nengo.Connection(const, a)
        nengo.Connection(const, b)
        nengo.Connection(const, c)
        ap = nengo.Probe(a.neurons, synapse=0.01)
        bp = nengo.Probe(b.neurons, synapse=0.01)
        cp = nengo.Probe(c.neurons, synapse=0.01)
    with Simulator(model) as sim:
        sim.run(0.06)
    t = sim.trange()

    plt.subplot(2, 1, 1)
    plt.plot(t, sim.data[ap], lw=3)
    plt.plot(t, sim.data[bp], lw=2)
    plt.plot(t, sim.data[cp])
    plt.subplot(2, 1, 2)
    plt.plot(*nengo.utils.ensemble.tuning_curves(a, sim), lw=3)
    plt.plot(*nengo.utils.ensemble.tuning_curves(b, sim), lw=2)
    plt.plot(*nengo.utils.ensemble.tuning_curves(c, sim))

    assert allclose(sim.data[ap], sim.data[bp])
    assert allclose(sim.data[bp], sim.data[cp])


def test_no_norm_encoders(Simulator, allclose):
    """Confirm encoders are not normalized"""

    enc_weight = 5
    with nengo.Network() as model:
        norm = nengo.Ensemble(1, 1, encoders=[[enc_weight]])
        no_norm = nengo.Ensemble(
            1, 1, encoders=[[enc_weight]], normalize_encoders=False
        )

    with Simulator(model) as sim:
        pass

    assert allclose(sim.data[norm].encoders, 1)
    assert allclose(sim.data[no_norm].encoders, enc_weight)


@pytest.mark.parametrize("intercept", [1.0, 1.1])
@pytest.mark.filterwarnings("ignore:divide by zero")
def test_raises_exception_for_invalid_intercepts(Simulator, intercept):
    with nengo.Network() as model:
        nengo.Ensemble(1, 1, intercepts=[intercept])

    with pytest.raises(BuildError):
        with Simulator(model):
            pass


def test_neurons_equality():
    with nengo.Network():
        neuron_type = nengo.LIF()
        ens0 = nengo.Ensemble(10, 1, neuron_type=neuron_type)
        assert ens0.neurons == ens0.neurons
        assert ens0.neurons != neuron_type

        ens1 = nengo.Ensemble(10, 1, neuron_type=neuron_type)
        assert ens0.neurons != ens1.neurons
        assert ens0.neuron_type == ens1.neuron_type

        # this is the only way you could ever have two neurons objects that are
        # equal but have different type
        old_neurons = ens1.neurons
        old_neuron_type = ens1.neuron_type
        ens1.neuron_type = nengo.RectifiedLinear()
        assert ens1.neurons == old_neurons
        assert ens1.neuron_type != old_neuron_type


def test_pickle(seed, Simulator):
    with nengo.Network(seed=seed) as net:
        net.ens = nengo.Ensemble(10, 1)
        net.probe = nengo.Probe(net.ens)

    with Simulator(net) as sim:
        sim.run(0.01)
    before = sim.data[net.probe]
    del sim

    pickled = pickle.dumps(net)
    del net
    unpickled = pickle.loads(pickled)

    with Simulator(unpickled) as sim:
        sim.run(0.01)
    after = sim.data[unpickled.probe]

    assert np.all(before == after)


def test_neurons_readonly():
    with nengo.Network():
        ens = nengo.Ensemble(10, 1)
        with pytest.raises(ReadonlyError, match="neurons"):
            ens.neurons = "test"


def test_probeable():
    with nengo.Network():
        ens = nengo.Ensemble(10, 1)
        assert ens.probeable == ("decoded_output", "input", "scaled_encoders")



# pylint: disable=wrong-import-order,wrong-import-position

"""Nengo provides a package for doing large-scale brain modelling in Python.

The source code repository for this package is found at
https://www.github.com/nengo/nengo. Examples of models can be found
in the `examples` directory of the source code repository.
"""

import logging
import sys

# Nengo namespace (API)
from nengo import spa  # pylint: disable=cyclic-import
from nengo import dists, exceptions, networks, presets, processes, utils
from nengo.base import Process
from nengo.config import Config
from nengo.connection import Connection
from nengo.ensemble import Ensemble
from nengo.learning_rules import BCM, PES, RLS, Oja, Voja
from nengo.network import Network
from nengo.neurons import (
    LIF,
    AdaptiveLIF,
    AdaptiveLIFRate,
    Direct,
    Izhikevich,
    LIFRate,
    PoissonSpiking,
    RectifiedLinear,
    RegularSpiking,
    Sigmoid,
    SpikingRectifiedLinear,
    StochasticSpiking,
    Tanh,
)
from nengo.node import Node
from nengo.params import Default
from nengo.probe import Probe
from nengo.rc import RC_DEFAULTS, rc
from nengo.simulator import Simulator
from nengo.synapses import Alpha, LinearFilter, Lowpass, Triangle
from nengo.transforms import (
    Conv,
    Convolution,
    ConvolutionTranspose,
    ConvTranspose,
    Dense,
    Sparse,
)
from nengo.version import version as __version__

logger = logging.getLogger(__name__)
try:
    # Prevent output if no handler set
    logger.addHandler(logging.NullHandler())
except AttributeError:  # pragma: no cover
    pass

__copyright__ = "2013-2021, Applied Brain Research"
__license__ = "Free for non-commercial use; see LICENSE.rst"



import numpy as np

from nengo import spa


def test_fixed(allclose):
    with spa.SPA() as model:
        model.buffer1 = spa.Buffer(dimensions=16)
        model.buffer2 = spa.Buffer(dimensions=8, subdimensions=2)
        model.input = spa.Input(buffer1="A", buffer2="B")

    input1, vocab1 = model.get_module_input("buffer1")
    input2, vocab2 = model.get_module_input("buffer2")

    assert allclose(model.input.input_nodes["buffer1"].output, vocab1.parse("A").v)
    assert allclose(model.input.input_nodes["buffer2"].output, vocab2.parse("B").v)


def test_time_varying(allclose):
    with spa.SPA() as model:
        model.buffer = spa.Buffer(dimensions=16)
        model.buffer2 = spa.Buffer(dimensions=16)

        def input(t):
            if t < 0.1:
                return "A"
            elif t < 0.2:
                return "B"
            else:
                return "0"

        model.input = spa.Input(buffer=input, buffer2="B")

    input, vocab = model.get_module_input("buffer")

    assert allclose(model.input.input_nodes["buffer"].output(t=0), vocab.parse("A").v)
    assert allclose(model.input.input_nodes["buffer"].output(t=0.1), vocab.parse("B").v)
    assert allclose(model.input.input_nodes["buffer"].output(t=0.2), np.zeros(16))


def test_predefined_vocabs(allclose):
    D = 64

    with spa.SPA() as model:
        model.vocab1 = spa.Vocabulary(D)
        model.vocab1.parse("A+B+C")

        model.vocab2 = spa.Vocabulary(D)
        model.vocab1.parse("A+B+C")

        model.buffer1 = spa.Buffer(dimensions=D, vocab=model.vocab1)
        model.buffer2 = spa.Buffer(dimensions=D, vocab=model.vocab2)

        def input(t):
            if t < 0.1:
                return "A"
            elif t < 0.2:
                return "B"
            else:
                return "C"

        model.input = spa.Input(buffer1=input, buffer2=input)

    a1 = model.input.input_nodes["buffer1"].output(t=0.0)
    b1 = model.input.input_nodes["buffer1"].output(t=0.1)
    c1 = model.input.input_nodes["buffer1"].output(t=0.2)

    a2 = model.input.input_nodes["buffer2"].output(t=0.0)
    b2 = model.input.input_nodes["buffer2"].output(t=0.1)
    c2 = model.input.input_nodes["buffer2"].output(t=0.2)

    assert allclose(a1, model.vocab1.parse("A").v)
    assert allclose(b1, model.vocab1.parse("B").v)
    assert allclose(c1, model.vocab1.parse("C").v)

    assert allclose(a2, model.vocab2.parse("A").v)
    assert allclose(b2, model.vocab2.parse("B").v)
    assert allclose(c2, model.vocab2.parse("C").v)

    assert np.dot(a1, a2) < 0.95
    assert np.dot(b1, b2) < 0.95
    assert np.dot(c1, c2) < 0.95



import numpy as np

import nengo
from nengo.exceptions import ValidationError
from nengo.spa.action_objects import DotProduct, Source
from nengo.spa.module import Module
from nengo.utils.numpy import is_number


class BasalGanglia(nengo.networks.BasalGanglia, Module):
    """A basal ganglia, performing action selection on a set of given actions.

    See `.networks.BasalGanglia` for more details.

    Parameters
    ----------
    actions : Actions
        The actions to choose between.
    input_synapse : float, optional
        The synaptic filter on all input connections.
    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.
    add_to_container : bool, optional
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(
        self, actions, input_synapse=0.002, label=None, seed=None, add_to_container=None
    ):
        self.actions = actions
        self.input_synapse = input_synapse
        self.spa = None
        self._bias = None
        Module.__init__(self)
        nengo.networks.BasalGanglia.__init__(
            self,
            dimensions=self.actions.count,
            label=label,
            seed=seed,
            add_to_container=add_to_container,
        )

    @property
    def bias(self):
        """Create a bias node, when needed."""
        if self._bias is None:
            with self:
                self._bias = nengo.Node([1], label="basal ganglia bias")
        return self._bias

    def on_add(self, spa):
        """Form the connections into the BG to compute the utility values.

        Each action's condition variable contains the set of computations
        needed for that action's utility value, which is the input to the
        basal ganglia.
        """
        Module.on_add(self, spa)
        self.spa = spa

        self.actions.process(spa)  # parse the actions

        for i, action in enumerate(self.actions.actions):
            cond = action.condition.expression
            # the basal ganglia hangles the condition part of the action;
            # the effect is handled by the thalamus

            # Note: A Source is an output from a module, and a Symbol is
            # text that can be parsed to be a SemanticPointer

            for c in cond.items:
                if isinstance(c, DotProduct):
                    if (isinstance(c.item1, Source) and c.item1.inverted) or (
                        isinstance(c.item2, Source) and c.item2.inverted
                    ):
                        raise NotImplementedError(
                            "Inversion in subexpression '%s' from action '%s' "
                            "is not supported by the Basal Ganglia." % (c, action)
                        )
                    if isinstance(c.item1, Source):
                        if isinstance(c.item2, Source):
                            # dot product between two different sources
                            self.add_compare_input(i, c.item1, c.item2, c.scale)
                        else:
                            self.add_dot_input(i, c.item1, c.item2, c.scale)
                    else:
                        # enforced in DotProduct constructor
                        assert isinstance(c.item2, Source)
                        self.add_dot_input(i, c.item2, c.item1, c.scale)
                elif isinstance(c, Source):
                    self.add_scalar_input(i, c)
                elif is_number(c):
                    self.add_bias_input(i, c)
                else:
                    raise NotImplementedError(
                        "Subexpression '%s' from action '%s' is not supported "
                        "by the Basal Ganglia." % (c, action)
                    )

    def add_bias_input(self, index, value):
        """Make an input that is just a fixed scalar value.

        Parameters
        ----------
        index : int
            the index of the action
        value : float or int
            the fixed utility value to add
        """
        with self.spa:
            nengo.Connection(
                self.bias,
                self.input[index : index + 1],
                transform=value,
                synapse=self.input_synapse,
            )

    def add_compare_input(self, index, source1, source2, scale):
        """Make an input that is the dot product of two different sources.

        This would be used for an input action such as ``dot(vision, memory)``.
        Each source might be transformed before being compared. If the
        two sources have different vocabularies, we use the vocabulary of
        the first one for comparison.

        Parameters
        ----------
        index : int
            The index of the action.
        source1 : Source
            The first module output to read from.
        source2 : Source
            The second module output to read from.
        scale : float
            A scaling factor to be applied to the result.
        """
        raise NotImplementedError(
            "Compare between two sources will never be "
            "implemented as discussed in "
            "https://github.com/nengo/nengo/issues/759"
        )

    def add_dot_input(self, index, source, symbol, scale):
        """Make an input that is the dot product of a Source and a Symbol.

        This would be used for an input action such as ``dot(vision, A)``.
        The source may have a transformation applied first.

        Parameters
        ----------
        index : int
            The index of the action.
        source : Source
            The module output to read from.
        symbol : Source
            The semantic pointer to compute the dot product with.
        scale : float
            A scaling factor to be applied to the result.
        """
        output, vocab = self.spa.get_module_output(source.name)
        # the first transformation, to handle dot(vision*A, B)
        t1 = vocab.parse(source.transform.symbol).get_convolution_matrix()
        # the linear transform to compute the fixed dot product
        t2 = np.array([vocab.parse(symbol.symbol).v * scale])

        transform = np.dot(t2, t1)

        with self.spa:
            nengo.Connection(
                output,
                self.input[index : index + 1],
                transform=transform,
                synapse=self.input_synapse,
            )

    def add_scalar_input(self, index, source):
        """Add a scalar input that will vary over time.

        This is used for the output of the `.Compare` module.

        Parameters
        ----------
        index : int
            The index of the action.
        source : Source
            The module output to read from.
        """
        output, _ = self.spa.get_module_output(source.name)
        if output.size_out != 1:
            raise NotImplementedError("Only 1-dimensional sources can be scalar inputs")

        try:
            scale = float(eval(source.transform.symbol))
        except ValueError:
            raise ValidationError(
                "Transform must be scalar; got '%s'" % source.transform.symbol,
                attr="source.transform",
            )

        with self.spa:
            nengo.Connection(
                output,
                self.input[index : index + 1],
                transform=scale,
                synapse=self.input_synapse,
            )



import sys

import numpy as np
import pytest

import nengo
import nengo.utils.numpy as npext
from nengo.base import Process
from nengo.dists import Distribution, Gaussian
from nengo.exceptions import ValidationError
from nengo.processes import (
    BrownNoise,
    FilteredNoise,
    Piecewise,
    PresentInput,
    WhiteNoise,
    WhiteSignal,
)
from nengo.synapses import Lowpass


class DistributionMock(Distribution):
    def __init__(self, retval):
        super().__init__()
        self.retval = retval
        self.sample_calls = []

    def sample(self, n, d=None, rng=np.random):
        self.sample_calls.append((n, d, rng))
        return np.ones((n, d)) * self.retval


class TimeProcess(Process):
    def make_step(self, shape_in, shape_out, dt, rng, state):
        size_in = np.prod(shape_in)
        size_out = np.prod(shape_out)
        if size_in == 0:
            return lambda t: [t] * size_out
        return lambda t, x: [t * np.sum(x)] * size_out


class QueueProcess(Process):
    def __init__(self):
        super().__init__()
        self.queue = []

    def make_step(self, shape_in, shape_out, dt, rng, state):
        def step(t, x, queue=self.queue):  # pylint: disable=dangerous-default-value
            queue.append(x)

        return step


def test_time(Simulator, allclose):
    t_run = 1.0
    c = 2.0
    process = TimeProcess()

    with nengo.Network() as model:
        u = nengo.Node(process)
        up = nengo.Probe(u)

        q = nengo.Node(c)
        v = nengo.Node(process, size_in=1, size_out=1)
        nengo.Connection(q, v, synapse=None)
        vp = nengo.Probe(v)

    with Simulator(model) as sim:
        sim.run(t_run)

    nt = len(sim.trange())
    assert process.trange(t_run).shape == sim.trange().shape
    assert allclose(process.trange(t_run), sim.trange())
    assert allclose(process.ntrange(nt), sim.trange())
    assert allclose(process.run(t_run), sim.data[up])
    assert allclose(process.apply([c] * nt), sim.data[vp])


def test_whitenoise(rng):
    dist = DistributionMock(42)
    process = WhiteNoise(dist, scale=False)
    samples = process.run_steps(5, d=2, rng=rng)
    assert np.all(samples == 42 * np.ones((5, 2)))
    assert process.run_steps(5, rng=rng).shape == (5, 1)
    assert process.run_steps(1, d=1, rng=rng).shape == (1, 1)
    assert process.run_steps(2, d=3, rng=rng).shape == (2, 3)


def test_brownnoise(Simulator, seed, plt):
    d = 5000
    t = 0.5
    std = 1.5
    process = BrownNoise(dist=Gaussian(0, std))
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(t)
    samples = sim.data[up]

    trange = sim.trange()
    expected_std = std * np.sqrt(trange)
    atol = 3.5 * expected_std / np.sqrt(d)

    plt.subplot(2, 1, 1)
    plt.title("Five Brown noise signals")
    plt.plot(trange, samples[:, :5])
    plt.subplot(2, 1, 2)
    plt.ylabel("Standard deviation")
    plt.plot(trange, np.abs(np.std(samples, axis=1)), label="Actual")
    plt.plot(trange, expected_std, label="Expected")
    plt.legend(loc="best")

    assert np.all(np.abs(np.mean(samples, axis=1)) < atol)
    assert np.all(np.abs(np.std(samples, axis=1) - expected_std) < atol)


def psd(values, dt=0.001):
    freq = npext.rfftfreq(values.shape[0], d=dt)
    power = (
        2.0
        * np.std(np.abs(np.fft.rfft(values, axis=0)), axis=1)
        / np.sqrt(values.shape[0])
    )
    return freq, power


@pytest.mark.parametrize("rms", [0.5, 1, 100])
def test_gaussian_whitenoise(Simulator, rms, seed, plt, allclose):
    d = 800
    process = WhiteNoise(Gaussian(0.0, rms), scale=False)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(0.3)
    values = sim.data[up]
    freq, val_psd = psd(values, dt=sim.dt)

    trange = sim.trange()
    plt.subplot(2, 1, 1)
    plt.title(f"First two dimensions of white noise process, rms={rms:.1f}")
    plt.plot(trange, values[:, :2])
    plt.xlim(right=trange[-1])
    plt.subplot(2, 1, 2)
    plt.title("Power spectrum")
    plt.plot(freq, val_psd, drawstyle="steps")

    val_rms = npext.rms(values, axis=0)
    assert allclose(val_rms.mean(), rms, rtol=0.02)
    assert allclose(val_psd[1:-1], rms, rtol=0.2)


@pytest.mark.parametrize("rms", [0.5, 1, 100])
def test_whitesignal_rms(Simulator, rms, seed, plt, allclose):
    t = 1.0
    d = 500
    process = WhiteSignal(t, high=500, rms=rms)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(t)
    values = sim.data[up]
    freq, val_psd = psd(values, dt=sim.dt)

    trange = sim.trange()
    plt.subplot(2, 1, 1)
    plt.title(f"First two D of white noise process, rms={rms:.1f}")
    plt.plot(trange, values[:, :2])
    plt.xlim(right=trange[-1])
    plt.subplot(2, 1, 2)
    plt.title("Power spectrum")
    plt.plot(freq, val_psd, drawstyle="steps")

    assert allclose(np.std(values), rms, rtol=0.02)
    assert allclose(val_psd[1:-1], rms, rtol=0.35)


@pytest.mark.parametrize("y0,d", [(0, 1), (-0.3, 3), (0.4, 1)])
def test_whitesignal_y0(Simulator, seed, y0, d):
    t = 0.1
    process = WhiteSignal(t, high=500, y0=y0)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(t)
    values = sim.data[up]
    error = np.min(abs(y0 - values), axis=0)

    assert ((y0 - error <= values[0, :]) & (values[0, :] <= y0 + error)).all()


@pytest.mark.parametrize("high,dt", [(10, 0.01), (5, 0.001), (50, 0.001)])
def test_whitesignal_high_dt(Simulator, high, dt, seed, plt, allclose):
    t = 1.0
    rms = 0.5
    d = 500
    process = WhiteSignal(t, high, rms=rms)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed, dt=dt) as sim:
        sim.run(t)
    values = sim.data[up]
    freq, val_psd = psd(values, dt=dt)

    trange = sim.trange()
    plt.subplot(2, 1, 1)
    plt.title(f"First two D of white noise process, high={high} Hz")
    plt.plot(trange, values[:, :2])
    plt.xlim(right=trange[-1])
    plt.subplot(2, 1, 2)
    plt.title("Power spectrum")
    plt.plot(freq, val_psd, drawstyle="steps")
    plt.xlim(0, high * 2.0)

    assert allclose(np.std(values, axis=1), rms, rtol=0.15)
    assert np.all(val_psd[npext.rfftfreq(len(trange), dt) > high] < rms * 0.5)


@pytest.mark.parametrize("high,dt", [(501, 0.001), (500, 0.002)])
def test_whitesignal_high_errors(Simulator, dt, high, seed):
    """Check for errors if ``high`` is not between 1/period and nyquist frequency."""

    with pytest.raises(ValidationError, match="Make ``high >= 1. / period``"):
        process = WhiteSignal(period=10 * dt, high=9 * dt)

    process = WhiteSignal(1.0, high=high)
    with nengo.Network() as model:
        nengo.Node(process, size_out=1)

    with pytest.raises(ValidationError, match="High must not exceed the Nyquist"):
        Simulator(model, dt=dt, seed=seed)


def test_whitesignal_continuity(Simulator, seed, plt):
    """Test that WhiteSignal is continuous over multiple periods."""
    t = 1.0
    high = 10
    rms = 0.5
    process = WhiteSignal(t, high=high, rms=rms)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=1)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(4 * t)
    dt = sim.dt
    x = sim.data[up]

    plt.plot(sim.trange(), x)

    # tolerances approximated from derivatives of sine wave of highest freq
    safety_factor = 2.0
    a, f = np.sqrt(2) * rms, (2 * np.pi * high) * dt
    assert abs(np.diff(x, axis=0)).max() <= safety_factor * a * f
    assert abs(np.diff(x, n=2, axis=0)).max() <= safety_factor ** 2 * a * f ** 2


def test_sampling_shape():
    process = WhiteSignal(0.1, high=500)
    assert process.run_steps(1).shape == (1, 1)
    assert process.run_steps(5, d=1).shape == (5, 1)
    assert process.run_steps(1, d=2).shape == (1, 2)


def test_x_copy(Simulator, allclose):
    """Test that process `x` is copied internally.

    If it is not copied, all elements in `process.queue` will reference the same
    underlying array, and will all be equal to each other.
    """
    with nengo.Network() as model:
        u = nengo.Node(lambda t: [t, t + 2])
        process = QueueProcess()
        v = nengo.Node(process, size_in=2)
        nengo.Connection(u, v, synapse=None)

    with Simulator(model) as sim:
        sim.run(0.003)

    t = sim.trange()
    assert allclose(process.queue, np.column_stack([t, t + 2]))


def test_reset(Simulator, seed, allclose):
    t_run = 0.1

    with nengo.Network() as model:
        u = nengo.Node(WhiteNoise(Gaussian(0, 1), scale=False))
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(t_run)
        x = np.array(sim.data[up])
        sim.reset()
        sim.run(t_run)
        y = np.array(sim.data[up])

    assert x.shape == y.shape
    assert allclose(x, y)


def test_frozen():
    """Test attributes inherited from FrozenObject"""
    a = WhiteNoise(dist=Gaussian(0.3, 0.2))
    b = WhiteNoise(dist=Gaussian(0.3, 0.2))
    c = FilteredNoise(dist=Gaussian(0.3, 0.2), synapse=Lowpass(0.02))

    assert hash(a) == hash(a)
    assert hash(b) == hash(b)
    assert hash(c) == hash(c)

    assert a == b
    assert hash(a) == hash(b)
    assert a != c
    assert hash(a) != hash(c)  # not guaranteed, but highly likely
    assert b != c
    assert hash(b) != hash(c)  # not guaranteed, but highly likely

    with pytest.raises(ValueError):
        a.dist = Gaussian(0.3, 0.5)  # test that dist param is frozen
    with pytest.raises(ValueError):
        a.dist.std = 0.4  # test that dist object is frozen


def test_seed(Simulator, seed, allclose):
    with nengo.Network() as model:
        a = nengo.Node(WhiteSignal(0.1, high=100, seed=seed))
        b = nengo.Node(WhiteSignal(0.1, high=100, seed=seed + 1))
        c = nengo.Node(WhiteSignal(0.1, high=100))
        d = nengo.Node(WhiteNoise(seed=seed))
        e = nengo.Node(WhiteNoise())
        ap = nengo.Probe(a)
        bp = nengo.Probe(b)
        cp = nengo.Probe(c)
        dp = nengo.Probe(d)
        ep = nengo.Probe(e)

    with Simulator(model) as sim1:
        sim1.run(0.1)

    with Simulator(model) as sim2:
        sim2.run(0.1)

    tols = dict(atol=1e-7, rtol=1e-4)
    assert allclose(sim1.data[ap], sim2.data[ap], **tols)
    assert allclose(sim1.data[bp], sim2.data[bp], **tols)
    assert not allclose(
        sim1.data[cp], sim2.data[cp], record_rmse=False, print_fail=0, **tols
    )
    assert not allclose(
        sim1.data[ap], sim1.data[bp], record_rmse=False, print_fail=0, **tols
    )
    assert allclose(sim1.data[dp], sim2.data[dp], **tols)
    assert not allclose(
        sim1.data[ep], sim2.data[ep], record_rmse=False, print_fail=0, **tols
    )


def test_present_input(Simulator, rng, allclose):
    n = 5
    c, ni, nj = 3, 8, 8
    images = rng.normal(size=(n, c, ni, nj))
    pres_time = 0.1

    model = nengo.Network()
    with model:
        u = nengo.Node(PresentInput(images, pres_time))
        up = nengo.Probe(u)

    with Simulator(model) as sim:
        sim.run(1.0)

    t = sim.trange()
    i = (np.floor((t - sim.dt) / pres_time + 1e-7) % n).astype(int)
    y = sim.data[up].reshape(len(t), c, ni, nj)
    for k, [ii, image] in enumerate(zip(i, y)):
        assert allclose(image, images[ii], rtol=1e-4, atol=1e-7), (k, ii)


class TestPiecewise:
    def run_sim(self, data, interpolation, Simulator):
        process = Piecewise(data, interpolation=interpolation)

        with nengo.Network() as model:
            u = nengo.Node(process, size_out=process.default_size_out)
            up = nengo.Probe(u)

        with Simulator(model) as sim:
            sim.run(0.15)

        return sim.trange(), sim.data[up]

    def test_basic(self, Simulator, allclose):
        t, f = self.run_sim({0.05: 1, 0.1: 0}, "zero", Simulator)
        assert allclose(f[t == 0.001], [0.0])
        assert allclose(f[t == 0.025], [0.0])
        assert allclose(f[t == 0.049], [0.0])
        assert allclose(f[t == 0.05], [1.0])
        assert allclose(f[t == 0.075], [1.0])
        assert allclose(f[t == 0.1], [0.0])
        assert allclose(f[t == 0.15], [0.0])

    def test_lists_in_data(self, Simulator, allclose):
        t, f = self.run_sim({0.05: [1, 0], 0.1: [0, 1]}, "zero", Simulator)
        assert allclose(f[t == 0.001], [0.0, 0.0])
        assert allclose(f[t == 0.025], [0.0, 0.0])
        assert allclose(f[t == 0.049], [0.0, 0.0])
        assert allclose(f[t == 0.05], [1.0, 0.0])
        assert allclose(f[t == 0.075], [1.0, 0.0])
        assert allclose(f[t == 0.1], [0.0, 1.0])
        assert allclose(f[t == 0.15], [0.0, 1.0])

    def test_default_zero(self, allclose):
        process = Piecewise({0.05: 1, 0.1: 0})
        f = process.make_step(
            shape_in=(process.default_size_in,),
            shape_out=(process.default_size_out,),
            dt=process.default_dt,
            rng=None,
            state={},
        )
        assert allclose(f(-10), [0.0])
        assert allclose(f(0), [0.0])

    def test_invalid_key(self):
        data = {0.05: 1, 0.1: 0, "a": 0.2}
        with pytest.raises(ValidationError, match=r"Keys must be times \(floats or in"):
            Piecewise(data)

    def test_invalid_callable(self):
        def badcallable(t):
            raise RuntimeError()

        data = {0.05: 1, 0.1: badcallable}
        with pytest.raises(ValidationError, match="should return a numerical const"):
            Piecewise(data)

    def test_invalid_length(self):
        data = {0.05: [1, 0], 0.1: [1, 0, 0]}
        with pytest.raises(ValidationError):
            Piecewise(data)

        # check that scalars and length 1 arrays can be used interchangeably
        # (no validation error)
        Piecewise({0.05: [1], 0.1: 0})

    def test_invalid_interpolation_type(self):
        data = {0.05: 1, 0.1: 0}
        with pytest.raises(ValidationError):
            Piecewise(data, interpolation="not-interpolation")

    def test_fallback_to_zero(self, monkeypatch):
        # Emulate not having scipy in case we have scipy
        monkeypatch.setitem(sys.modules, "scipy.interpolate", None)

        with pytest.warns(UserWarning, match="cannot be applied.*scipy is not inst"):
            process = Piecewise({0.05: 1, 0.1: 0}, interpolation="linear")
        assert process.interpolation == "zero"

    def test_interpolation_1d(self, plt, Simulator, allclose):
        pytest.importorskip("scipy")

        # Note: cubic requires an explicit start of 0
        data = {0: -0.5, 0.05: 1, 0.075: -1, 0.1: -0.5}

        def test_and_plot(interp):
            t, f = self.run_sim(data, interp, Simulator)
            assert allclose(f[t == 0.05], [1.0])
            assert allclose(f[t == 0.075], [-1.0])
            assert allclose(f[t == 0.1], [-0.5])
            plt.plot(t, f, label=interp)

        test_and_plot("zero")
        test_and_plot("linear")
        test_and_plot("nearest")
        test_and_plot("slinear")
        test_and_plot("quadratic")
        test_and_plot("cubic")
        plt.legend(loc="lower left")

    def test_interpolation_2d(self, plt, Simulator, allclose):
        pytest.importorskip("scipy")

        # Note: cubic requires an explicit start of 0
        data = {
            0: [-0.5, -0.5],
            0.05: [1.0, 0.5],
            0.075: [-1, -0.5],
            0.1: [-0.5, -0.25],
        }

        ax1 = plt.subplot(2, 1, 1)
        ax2 = plt.subplot(2, 1, 2)

        def test_and_plot(interp):
            t, f = self.run_sim(data, interp, Simulator)
            assert allclose(f[t == 0.05], [1.0, 0.5])
            assert allclose(f[t == 0.075], [-1.0, -0.5])
            assert allclose(f[t == 0.1], [-0.5, -0.25])
            ax1.plot(t, f.T[0], label=interp)
            ax2.plot(t, f.T[1], label=interp)

        test_and_plot("zero")
        test_and_plot("linear")
        test_and_plot("nearest")
        test_and_plot("slinear")
        test_and_plot("quadratic")
        test_and_plot("cubic")
        plt.legend(loc="lower left")

    @pytest.mark.parametrize("value", (1, [1, 1]))
    def test_shape_out(self, value):
        process = Piecewise({1: value})
        f = process.make_step(
            shape_in=(process.default_size_in,),
            shape_out=(process.default_size_out,),
            dt=process.default_dt,
            rng=None,
            state={},
        )

        assert np.array_equal(f(0), np.zeros(process.default_size_out))
        assert np.array_equal(f(2), np.ones(process.default_size_out))

    def test_function(self, Simulator, allclose):
        t, f = self.run_sim({0.05: np.sin, 0.1: np.cos}, "zero", Simulator)
        assert allclose(f[t == 0.001], [0.0])
        assert allclose(f[t == 0.049], [0.0])
        assert allclose(f[t == 0.05], [np.sin(0.05)])
        assert allclose(f[t == 0.075], [np.sin(0.075)])
        assert allclose(f[t == 0.1], [np.cos(0.1)])
        assert allclose(f[t == 0.15], [np.cos(0.15)])

    def test_function_list(self, Simulator, allclose):
        def func1(t):
            return t, t ** 2, t ** 3

        def func2(t):
            return t ** 4, t ** 5, t ** 6

        t, f = self.run_sim({0.05: func1, 0.1: func2}, "zero", Simulator)
        assert allclose(f[t == 0.001], [0.0])
        assert allclose(f[t == 0.049], [0.0])
        assert allclose(f[t == 0.05], func1(0.05))
        assert allclose(f[t == 0.075], func1(0.075))
        assert allclose(f[t == 0.1], func2(0.1))
        assert allclose(f[t == 0.15], func2(0.15))

    def test_mixture(self, Simulator, allclose):
        t, f = self.run_sim({0.05: 1, 0.1: np.cos}, "zero", Simulator)
        assert allclose(f[t == 0.001], [0.0])
        assert allclose(f[t == 0.049], [0.0])
        assert allclose(f[t == 0.05], [1.0])
        assert allclose(f[t == 0.075], [1.0])
        assert allclose(f[t == 0.1], [np.cos(0.1)])
        assert allclose(f[t == 0.15], [np.cos(0.15)])

    def test_mixture_3d(self, Simulator, allclose):
        def func(t):
            return t, t ** 2, t ** 3

        t, f = self.run_sim({0.05: [1, 1, 1], 0.1: func}, "zero", Simulator)
        assert allclose(f[t == 0.001], [0.0, 0.0, 0.0])
        assert allclose(f[t == 0.049], [0.0, 0.0, 0.0])
        assert allclose(f[t == 0.05], [1.0, 1.0, 1.0])
        assert allclose(f[t == 0.075], [1.0, 1.0, 1.0])
        assert allclose(f[t == 0.1], func(0.1))
        assert allclose(f[t == 0.15], func(0.15))

    def test_invalid_function_length(self):
        with pytest.raises(ValidationError, match="time 1.0 has size 2"):
            Piecewise({0.5: 0, 1.0: lambda t: [t, t ** 2]})

    def test_invalid_interpolation_on_func(self):
        def func(t):
            return t

        with pytest.warns(UserWarning, match="cannot be applied.*callable was sup"):
            process = Piecewise({0.05: 0, 0.1: func}, interpolation="linear")
        assert process.interpolation == "zero"

    def test_cubic_interpolation_warning(self):
        pytest.importorskip("scipy")

        # cubic interpolation with 0 not in times
        process = Piecewise({0.001: 0, 0.1: 0.1}, interpolation="cubic")
        with pytest.warns(UserWarning, match="'cubic' interpolation.*for t=0.0"):
            try:
                process.run(0.001)
            except ValueError:
                pass  # scipy may raise a ValueError



# pylint: disable=missing-docstring

import ast
import sys

import pytest

from nengo.utils.paths import install_dir


@pytest.mark.skipif(
    sys.version_info < (3, 8, 0),
    reason="ast.parse `feature_version` added in Python 3.8",
)
@pytest.mark.parametrize("feature_version", [(3, 4), (3, 5)])
def test_setup_compat(feature_version):
    setup_py_path = install_dir / "setup.py"

    assert setup_py_path.exists()
    with setup_py_path.open("r") as fh:
        source = fh.read()

    parsed = ast.parse(source, feature_version=feature_version)
    assert parsed is not None



import numpy as np
import pytest

from nengo.builder.processes import SimProcess
from nengo.builder.tests.test_operator import _test_operator_arg_attributes
from nengo.processes import Process


def test_simprocess():
    argnames = ["process", "input", "output", "t"]
    non_signals = ["process"]
    _, sim = _test_operator_arg_attributes(
        SimProcess, argnames, non_signals=non_signals
    )
    assert str(sim) == "SimProcess{process, input -> output}"

    with pytest.raises(ValueError, match="Unrecognized mode"):
        _test_operator_arg_attributes(SimProcess, argnames, args={"mode": "badval"})


@pytest.mark.parametrize("mode", ["set", "inc"])
@pytest.mark.parametrize("has_input", [False, True])
def test_simprocess_make_step(mode, has_input, rng):
    t0 = rng.uniform(size=1)
    in0 = rng.uniform(size=1)
    out0 = rng.uniform(size=1)
    ref = t0 + (in0 if has_input else 0) + (out0 if mode == "inc" else 0)

    signals = {"in": in0.copy(), "out": out0.copy(), "t": t0.copy()}
    sim = SimProcess(
        TimeAddProcess(),
        input="in" if has_input else None,
        output="out",
        t="t",
        mode=mode,
    )
    step = sim.make_step(signals, dt=1, rng=rng)
    step()

    assert np.array_equal(signals["out"], ref)


class TimeAddProcess(Process):
    def make_step(self, shape_in, shape_out, dt, rng, state):
        return (lambda t: t) if np.prod(shape_in) == 0 else (lambda t, x: x + t)



"""Some generic decorators and a decorator for making better decorators.

Decorators implemented as function closures have important flaws:
    - __name__ and __doc__ are wrong
    - function argspec is wrong
    - function source code cannot be retrieved
    - cannot be applied on top of decorators implemented as descriptors

The ``decorator`` decorator in this file solves these problems.
``@decorator`` should be applied to all decorator functions in Nengo.
This makes decorators easy to write, while avoiding the issues above.

----

Some of the code in this file is a modified subset of the wrapt package,
available at https://pypi.org/project/wrapt/.
For more details on the implementations herein, see the code comments
in the wrapt source at https://github.com/GrahamDumpleton/wrapt.
The code is included here under the terms of wrapt's BSD license:

Copyright (c) 2013, Graham Dumpleton. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

import functools
import inspect


class ObjectProxyMethods:
    """Properly override __module__ and __doc__.

    If we add these in ObjectProxy, the derived class __dict__ will
    still be setup to have string variants of these attributes and the
    rules of descriptors means that they appear to take precedence
    over the properties in the base class. To avoid that, we copy the
    properties into the derived class type itself via a meta class. In
    that way the properties will always take precedence.
    """

    @property
    def __module__(self):  # pylint: disable=function-redefined
        return self.__wrapped__.__module__

    @property
    def __doc__(self):
        return self.__wrapped__.__doc__

    @property
    def __dict__(self):
        """Must be explicit to ensure that vars() works as expected."""
        return self.__wrapped__.__dict__


class ObjectProxyMeta(type):
    """Ensure ObjectProxyMethods take precedence.

    Copy our special properties into the class so that they always
    take precedence over attributes of the same name added during
    construction of a derived class. This is to save duplicating the
    implementation for them in all derived classes.
    """

    def __new__(metacls, name, bases, dictionary):
        dictionary.update(vars(ObjectProxyMethods))
        return type.__new__(metacls, name, bases, dictionary)


class ObjectProxy(metaclass=ObjectProxyMeta):
    """A transparent object proxy for creating decorator descriptors.

    This is used in lieu of ``functools.update_wrapper``, which copies
    a number of properties of the wrapped function in the wrapper.
    Copying properties can be expensive though, so this is used instead
    to make the wrapper act like the wrapped function in all cases
    except ``__call__``.
    """

    __slots__ = ["__wrapped__"]

    def __init__(self, wrapped):
        object.__setattr__(self, "__wrapped__", wrapped)

        # Python 3 has the __qualname__ attribute, but it does not
        # allow it to be overridden using a property and it must instead
        # be an actual string object instead.
        try:
            object.__setattr__(self, "__qualname__", wrapped.__qualname__)
        except AttributeError:
            pass

    @property
    def __annotations__(self):
        return self.__wrapped__.__annotations__

    @property
    def __name__(self):
        return self.__wrapped__.__name__

    @property
    def __class__(self):
        return type(self.__wrapped__)

    def __dir__(self):
        return dir(self.__wrapped__)

    def __getattr__(self, key):
        return getattr(self.__wrapped__, key)

    def __hash__(self):
        return hash(self.__wrapped__)

    def __setattr__(self, key, value):
        setattr(self.__wrapped__, key, value)

    def __str__(self):
        return str(self.__wrapped__)

    def __repr__(self):
        return (
            f"<{type(self).__name__} at 0x{id(self):x} "
            f"for {type(self.__wrapped__).__name__} at 0x{id(self.__wrapped__):x}>"
        )


class BoundFunctionWrapper(ObjectProxy):
    """A descriptor to emulate a bound function.

    This is used to create bound function decorators.
    It maintains all of the nice introspection that can normally
    be done on bound functions.
    """

    __slots__ = ("instance", "wrapper", "binding", "parent")

    def __init__(self, wrapped, instance, wrapper, binding, parent):
        super().__init__(wrapped)
        # Using object.__setattr__ to subvert ObjectProxy.__setattr__
        object.__setattr__(self, "instance", instance)
        object.__setattr__(self, "wrapper", wrapper)
        object.__setattr__(self, "binding", binding)
        object.__setattr__(self, "parent", parent)

    def __get__(self, instance, owner):
        if self.instance is None and self.binding == "function":
            descriptor = self.parent.__wrapped__.__get__(instance, owner)
            return BoundFunctionWrapper(
                descriptor, instance, self.wrapper, self.binding, self.parent
            )
        return self

    def __call__(self, *args, **kwargs):
        if self.binding == "function":
            if self.instance is None:
                instance, args = args[0], args[1:]
                wrapped = functools.partial(self.__wrapped__, instance)
                return self.wrapper(wrapped, instance, args, kwargs)
            else:
                return self.wrapper(self.__wrapped__, self.instance, args, kwargs)
        else:
            instance = getattr(self.__wrapped__, "__self__", None)
            return self.wrapper(self.__wrapped__, instance, args, kwargs)


class FunctionWrapper(ObjectProxy):
    """A descriptor to emulate a function.

    This is used to create function decorators.
    It maintains all of the nice introspection that can normally
    be done on functions.
    """

    __slots__ = ("wrapper", "binding")

    def __init__(self, wrapped, wrapper):
        super().__init__(wrapped)
        # Using object.__setattr__ to subvert ObjectProxy.__setattr__
        object.__setattr__(self, "wrapper", wrapper)
        if isinstance(wrapped, classmethod):
            object.__setattr__(self, "binding", "classmethod")
        elif isinstance(wrapped, staticmethod):
            object.__setattr__(self, "binding", "staticmethod")
        else:
            object.__setattr__(self, "binding", "function")

    def __get__(self, instance, owner):
        wrapped = self.__wrapped__.__get__(instance, owner)
        return BoundFunctionWrapper(wrapped, instance, self.wrapper, self.binding, self)

    def __call__(self, *args, **kwargs):
        return self.wrapper(self.__wrapped__, None, args, kwargs)


def decorator(wrapper):
    """Decorate decorators.

    This imposes a particular style of writing descriptors.
    The descriptor must accept four positional arguments:

    - ``wrapped``: the function being wrapped
    - ``instance``: the instance that is bound to the function in the case of
      bound functions (None in the case of plain functions)
    - ``args``: the positional arguments passed to the wrapped function
    - ``kwargs``: the keyword arguments passed to the wrapped function

    Examples
    --------

    Decorating a normal function (i.e., instance will always be None):

    .. testcode::

       from nengo.utils.magic import decorator

       @decorator
       def my_decorator(wrapped, instance, args, kwargs):
           return wrapped(*args, **kwargs)

       @my_decorator
       def f():
           return 1

    Decorating a bound function:

    .. testcode::

       @decorator
       def my_decorator(wrapped, instance, args, kwargs):
           return wrapped(*args, **kwargs)

       class MyClass:
           @my_decorator
           def f(self):
               return self.num + 1
    """

    def _wrapper(wrapped, instance, args, kwargs):
        def _execute(wrapped, *_args, **_kwargs):
            if instance is None:
                target_wrapper = wrapper
            elif inspect.isclass(instance):
                target_wrapper = wrapper.__get__(None, instance)
            else:
                target_wrapper = wrapper.__get__(instance, type(instance))
            return FunctionWrapper(wrapped, target_wrapper)

        return _execute(*args, **kwargs)

    return FunctionWrapper(wrapper, _wrapper)
